# coding: utf-8
"""utils function for esrf volumes"""

import os
import shutil

import numpy
import pytest

from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.esrf.volume.hdf5volume import HDF5Volume
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume, has_glymur, has_minimal_openjpeg
from tomoscan.esrf.volume.tiffvolume import MultiTIFFVolume, TIFFVolume, has_tifffile
from tomoscan.esrf.volume.utils import guess_hdf5_volume_data_paths, guess_volumes


@pytest.mark.parametrize("file_extension", ("h5", "hdf5"))
def test_guess_volumes_hdf5(tmp_path, file_extension):
    """
    Test `guess_volumes` function within hdf5 file
    """
    assert guess_volumes(path=tmp_path) == tuple()
    file_path = os.path.join(tmp_path, ".".join(["my_file", file_extension]))
    # check error if path does not exists
    with pytest.raises(OSError):
        guess_volumes(path=file_path)

    volume = HDF5Volume(
        file_path=file_path,
        data_path="entry",
        data=numpy.linspace(1, 10, 100).reshape(1, 10, 10),
    )
    volume.save()

    guess_vols = guess_volumes(path=file_path)
    assert len(guess_vols) == 1
    guess_vol = guess_vols[0]
    assert isinstance(guess_vol, HDF5Volume)
    assert guess_vol.file_path == file_path
    assert guess_vol.data_path == "/entry"

    HDF5Volume(
        file_path=file_path,
        data_path="entry0002",
        data=numpy.linspace(1, 10, 100).reshape(1, 10, 10),
    ).save()
    assert len(guess_volumes(path=file_path)) == 2

    # a few more test on 'guess_hdf5_volume_data_paths'
    assert len(guess_hdf5_volume_data_paths(file_path=file_path, depth=0)) == 0
    assert (
        len(
            guess_hdf5_volume_data_paths(
                file_path=file_path, data_path="/entry", depth=0
            )
        )
        == 1
    )


_constructors = [EDFVolume]
if has_glymur and has_minimal_openjpeg:
    _constructors.append(JP2KVolume)
if has_tifffile:
    _constructors.append(TIFFVolume)


@pytest.mark.parametrize("Constructor", _constructors)
def test_guess_volumes_single_frame_file(tmp_path, Constructor):
    """
    Test `guess_volumes` function for single frame volume (like EDFVolume, TIFFVolume and Jp2KVolume)
    """
    assert guess_volumes(path=tmp_path) == tuple()
    folder = os.path.join(tmp_path, "my_volume_folder")
    # check error if path does not exists
    with pytest.raises(OSError):
        guess_volumes(path=folder)

    volume = Constructor(
        folder=folder,
        data=numpy.linspace(1, 10, 30000, dtype=numpy.uint8).reshape(3, 100, 100),
    )
    volume.save()

    guess_vols = guess_volumes(path=folder)
    assert len(guess_vols) == 1
    guess_vol = guess_vols[0]
    assert isinstance(guess_vol, Constructor)
    assert guess_vol.data_url.file_path() == folder
    assert guess_vol.get_volume_basename() == os.path.basename(folder)
    # test adding some noise in the folder
    for _ in range(15):
        file_name = os.path.join(folder, "my_volume_folder.txt")
        with open(file_name, mode="w") as f:
            f.write("almost empty text file")
    for other_file in ("other_file.tif", "other_file.tiff"):
        with open(os.path.join(folder, other_file), mode="w") as f:
            f.write("none")

    guess_vols = guess_volumes(path=folder)
    assert len(guess_vols) == 1
    shutil.rmtree(folder)
    # test with a basename
    volume = Constructor(
        folder=folder,
        volume_basename="other_basename",
        data=numpy.linspace(1, 10, 10000, dtype=numpy.uint8).reshape(1, 100, 100),
    )
    volume.save()
    guess_vols = guess_volumes(path=folder)
    assert len(guess_vols) == 1
    assert guess_vols[0].get_volume_basename() == "other_basename"


@pytest.mark.skipif(not has_tifffile, reason="tiffile not available")
def test_guess_volumes_multitiff(tmp_path):
    """
    Test `guess_volumes` function for MultiTiffVolume
    """
    assert guess_volumes(path=tmp_path) == tuple()
    file_path = os.path.join(tmp_path, "my_multitiff.tiff")
    # check error if path does not exists
    with pytest.raises(OSError):
        guess_volumes(path=file_path)

    volume = MultiTIFFVolume(
        file_path=file_path, data=numpy.linspace(1, 10, 200).reshape(2, 10, 10)
    )
    volume.save()

    guess_vols = guess_volumes(path=file_path)
    assert len(guess_vols) == 1
    guess_vol = guess_vols[0]
    assert isinstance(guess_vol, MultiTIFFVolume)
    assert guess_vol.data_url.file_path() == file_path



# coding: utf-8

import os
import shutil
import tempfile
import unittest

import h5py
import numpy
import pytest
import pint
from silx.io.url import DataUrl
from silx.io.utils import get_data

from tomoscan.esrf.scan.nxtomoscan import NXtomoScan, ImageKey, TomoFrame
from tomoscan.esrf.scan.mock import MockNXtomo
from tomoscan.nexus.paths.nxtomo import get_paths as get_nexus_paths
from tomoscan.nexus.paths.nxtomo import nx_tomo_path_latest, nx_tomo_path_v_1_0
from tomoscan.scanbase import ReducedFramesInfos
from tomoscan.tests.utils import NXtomoMockContext
from tomoscan.tests.datasets import GitlabDataset
from nxtomo.application.nxtomo import NXtomo

_ureg = pint.get_application_registry()


class HDF5TestBaseClass(unittest.TestCase):
    """base class for hdf5 unit test"""

    def get_dataset(self, hdf5_dataset_name):
        output_dataset_file = os.path.join(
            self.test_dir, os.path.basename(hdf5_dataset_name)
        )
        raw_dataset_file = GitlabDataset.get_dataset(hdf5_dataset_name)
        if raw_dataset_file is None:
            raise ValueError(f"Unable to retrieve {hdf5_dataset_name}")
        shutil.copy(src=raw_dataset_file, dst=output_dataset_file)
        return output_dataset_file

    def setUp(self) -> None:
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self) -> None:
        shutil.rmtree(self.test_dir)


class TestHDF5Scan(HDF5TestBaseClass):
    """Basic test for the hdf5 scan"""

    def setUp(self) -> None:
        super(TestHDF5Scan, self).setUp()
        self.dataset_file = self.get_dataset(
            "h5_datasets/frm_edftomomill_twoentries.nx"
        )
        assert os.path.exists(self.dataset_file)
        self.scan = NXtomoScan(scan=self.dataset_file)
        self.scan.nexus_version = 1.0

    def testGeneral(self):
        """some general on the HDF5Scan"""
        self.assertEqual(self.scan.master_file, self.dataset_file)
        self.assertEqual(self.scan.path, os.path.dirname(self.dataset_file))
        self.assertEqual(self.scan.type, "hdf5")
        self.assertEqual(self.scan.entry, "/entry0000")
        self.assertEqual(len(self.scan.flats), 42)
        self.assertEqual(len(self.scan.darks), 1)
        self.assertEqual(
            len(self.scan.return_projs), 0
        )  # because frm_edftomomill_twoentries doesn't contain the image_key_control dataset. So in this case we always have 0 alignment projections
        self.assertEqual(self.scan.tomo_n, None)
        self.assertEqual(self.scan.start_time, None)
        self.assertEqual(self.scan.end_time, None)
        self.assertEqual(self.scan.energy, 17.05)
        self.assertEqual(self.scan.field_of_view, None)
        self.assertEqual(self.scan.x_rotation_axis_pixel_position, None)
        self.assertEqual(len(self.scan.x_translation), 1546)
        self.assertEqual(len(self.scan.y_translation), 1546)
        self.assertEqual(len(self.scan.z_translation), 1546)

        proj_angles = self.scan.get_proj_angle_url()
        self.assertEqual(
            len(proj_angles), 1500 + 3 - 1
        )  # -1 because the dataset has twice the same example defined. Has there is no image_key_control return projection are not handled.
        self.assertTrue(90 in proj_angles)
        self.assertTrue(24.0 in proj_angles)
        self.assertTrue(179.88 not in proj_angles)

        url_1 = proj_angles[0]
        self.assertTrue(url_1.is_valid())
        self.assertTrue(url_1.is_absolute())
        self.assertEqual(url_1.scheme(), "silx")
        # check conversion to dict
        _dict = self.scan.to_dict()
        scan2 = NXtomoScan.from_dict(_dict)
        self.assertEqual(scan2.master_file, self.scan.master_file)
        self.assertEqual(scan2.entry, self.scan.entry)

    def testFrames(self):
        """Check the `frames` property which is massively used under the
        NXtomoScan class"""
        frames = self.scan.frames
        # check some projections
        proj_2 = frames[24]
        self.assertTrue(isinstance(proj_2, TomoFrame))
        self.assertEqual(proj_2.index, 24)
        numpy.isclose(proj_2.rotation_angle, 0.24)
        self.assertFalse(proj_2.is_control)
        self.assertEqual(proj_2.url.file_path(), self.scan.master_file)
        self.assertEqual(proj_2.url.data_path(), "/entry0000/instrument/detector/data")
        self.assertEqual(proj_2.url.data_slice(), 24)
        self.assertEqual(proj_2.image_key, ImageKey.PROJECTION)
        self.assertEqual(get_data(proj_2.url).shape, (20, 20))
        # check last two non-return projection
        for frame_index in (1520, 1542):
            with self.subTest(frame_index=frame_index):
                frame = frames[frame_index]
                self.assertTrue(frame.image_key, ImageKey.PROJECTION)
                self.assertFalse(frame.is_control)

        # check some darks
        dark_0 = frames[0]
        self.assertEqual(dark_0.index, 0)
        numpy.isclose(dark_0.rotation_angle, 0.0)
        self.assertFalse(dark_0.is_control)
        self.assertEqual(dark_0.url.file_path(), self.scan.master_file)
        self.assertEqual(dark_0.url.data_path(), "/entry0000/instrument/detector/data")
        self.assertEqual(dark_0.url.data_slice(), 0)
        self.assertEqual(dark_0.image_key, ImageKey.DARK_FIELD)
        self.assertEqual(get_data(dark_0.url).shape, (20, 20))

        # check some flats
        flat_1 = frames[2]
        self.assertEqual(flat_1.index, 2)
        numpy.isclose(flat_1.rotation_angle, 0.0)
        self.assertFalse(flat_1.is_control)
        self.assertEqual(flat_1.url.file_path(), self.scan.master_file)
        self.assertEqual(flat_1.url.data_path(), "/entry0000/instrument/detector/data")
        self.assertEqual(flat_1.url.data_slice(), 2)
        self.assertEqual(flat_1.image_key, ImageKey.FLAT_FIELD)
        self.assertEqual(get_data(flat_1.url).shape, (20, 20))

        # check some return projections
        r_proj_0 = frames[1543]
        self.assertTrue(isinstance(r_proj_0, TomoFrame))
        self.assertEqual(r_proj_0.index, 1543)
        numpy.isclose(r_proj_0.rotation_angle, 180)
        self.assertFalse(
            r_proj_0.is_control
        )  # there is no image_key_control dataset in the current dataset
        self.assertEqual(r_proj_0.url.file_path(), self.scan.master_file)
        self.assertEqual(
            r_proj_0.url.data_path(), "/entry0000/instrument/detector/data"
        )
        self.assertEqual(r_proj_0.url.data_slice(), 1543)
        self.assertEqual(r_proj_0.image_key, ImageKey.PROJECTION)
        self.assertEqual(get_data(r_proj_0.url).shape, (20, 20))

    def testProjections(self):
        """Make sure projections are valid"""
        projections = self.scan.projections
        self.assertEqual(len(self.scan.projections), 1503)
        url_0 = projections[list(projections.keys())[0]]
        self.assertEqual(url_0.file_path(), os.path.join(self.scan.master_file))
        self.assertEqual(url_0.data_slice(), 22)
        # should be 4 but angles are truely missleading: 179.88, 180.0, 90, 0.
        # in this case we are not using any information from image_key_control
        # and we wait deduce 'return mode' from angles.
        self.assertEqual(
            len(self.scan.alignment_projections), 0
        )  # frm_edftomomill_twoentries doesn't contain the image_key_control

    def testDark(self):
        """Make sure darks are valid"""
        n_dark = 1
        self.assertEqual(self.scan.dark_n, n_dark)
        darks = self.scan.darks
        self.assertEqual(len(darks), 1)
        # TODO check accumulation time

    def testFlats(self):
        """Make sure flats are valid"""
        n_flats = 42
        flats = self.scan.flats
        self.assertEqual(len(flats), n_flats)
        self.assertEqual(
            self.scan.flat_n, n_flats // 2
        )  # because get two series of flat
        with self.assertRaises(NotImplementedError):
            self.scan.ff_interval

    def testDims(self):
        self.assertEqual(self.scan.dim_1, 20)
        self.assertEqual(self.scan.dim_2, 20)

    def testAxisUtils(self):
        self.assertEqual(self.scan.scan_range, 180)
        self.assertEqual(len(self.scan.projections), 1503)

        radios_urls_evolution = self.scan.get_proj_angle_url()
        self.assertEqual(
            len(radios_urls_evolution), 1502
        )  # get 1502 because there is one overwrite because get twice the same exact position
        self.assertEqual(radios_urls_evolution[0].file_path(), self.scan.master_file)
        self.assertEqual(radios_urls_evolution[0].data_slice(), 22)
        self.assertEqual(
            radios_urls_evolution[0].data_path(), "/entry0000/instrument/detector/data"
        )

    def testDarkRefUtils(self):
        self.assertEqual(len(self.scan.projections), 1503)
        pixel_size = self.scan.pixel_size
        self.assertTrue(pixel_size is not None)
        self.assertTrue(
            numpy.isclose(self.scan.pixel_size * _ureg.meter, 0.05 * _ureg.millimeter)
        )
        self.assertTrue(numpy.isclose(self.scan.get_pixel_size(unit="micrometer"), 50))
        self.assertTrue(
            numpy.isclose(
                self.scan.sample_x_pixel_size * _ureg.meter,
                0.05 * _ureg.millimeter,
            )
        )
        self.assertTrue(
            numpy.isclose(
                self.scan.sample_y_pixel_size * _ureg.meter,
                0.05 * _ureg.millimeter,
            )
        )

    def testNabuUtil(self):
        self.assertTrue(numpy.isclose(self.scan.sample_detector_distance, -19.9735))
        self.assertTrue(
            numpy.isclose(self.scan.get_sample_detector_distance(unit="cm"), -1997.35)
        )

    def testCompactedProjs(self):
        projs_compacted = self.scan.projections_compacted
        self.assertEqual(projs_compacted.keys(), self.scan.projections.keys())
        for i in range(22, 1520 + 1):
            self.assertEqual(projs_compacted[i].data_slice(), slice(22, 1521, None))
        for i in range(1542, 1543):
            self.assertEqual(projs_compacted[i].data_slice(), slice(1542, 1546, None))


class TestFlatFieldCorrection(HDF5TestBaseClass):
    """Test the flat field correction"""

    def setUp(self) -> None:
        super(TestFlatFieldCorrection, self).setUp()
        self.dataset_file = self.get_dataset(
            "h5_datasets/frm_edftomomill_twoentries.nx"
        )
        self.scan = NXtomoScan(scan=self.dataset_file)

    def testFlatAndDarksSet(self):
        self.scan.set_reduced_flats(
            {
                21: numpy.random.random(20 * 20).reshape((20, 20)),
                1520: numpy.random.random(20 * 20).reshape((20, 20)),
            }
        )
        self.scan.set_reduced_darks({0: numpy.random.random(20 * 20).reshape((20, 20))})

        projs = []
        proj_indexes = []
        for proj_index, proj in self.scan.projections.items():
            projs.append(proj)
            proj_indexes.append(proj_index)
        normed_proj = self.scan.flat_field_correction(
            projs=projs, proj_indexes=proj_indexes
        )
        self.assertEqual(len(normed_proj), len(self.scan.projections))
        raw_data = get_data(projs[50])
        self.assertFalse(numpy.array_equal(raw_data, normed_proj[50]))

    def testNoFlatOrDarkSet(self):
        projs = []
        proj_indexes = []
        for proj_index, proj in self.scan.projections.items():
            projs.append(proj)
            proj_indexes.append(proj_index)
        with self.assertLogs("tomoscan", level="ERROR"):
            normed_proj = self.scan.flat_field_correction(
                projs=projs, proj_indexes=proj_indexes
            )
        self.assertEqual(len(normed_proj), len(self.scan.projections))
        raw_data = get_data(projs[50])
        self.assertTrue(numpy.array_equal(raw_data, normed_proj[50]))


class TestGetSinogram(HDF5TestBaseClass):
    """Test the get_sinogram function"""

    def setUp(self) -> None:
        super(TestGetSinogram, self).setUp()
        self.dataset_file = self.get_dataset(
            "h5_datasets/frm_edftomomill_twoentries.nx"
        )

        self.scan = NXtomoScan(scan=self.dataset_file)
        # set some random dark and flats
        self.scan.set_reduced_flats(
            {
                21: numpy.random.random(20 * 20).reshape((20, 20)),
                1520: numpy.random.random(20 * 20).reshape((20, 20)),
            }
        )
        dark = numpy.random.random(20 * 20).reshape((20, 20))
        self.scan.set_reduced_darks({0: dark})
        self.scan._flats_weights = self.scan._get_flats_weights()
        self._raw_frame = []
        for index, url in self.scan.projections.items():
            self._raw_frame.append(get_data(url))
        self._raw_frame = numpy.asarray(self._raw_frame)

        assert self._raw_frame.ndim == 3

        normed_frames = []
        for proj_i, z_frame in enumerate(self._raw_frame):
            normed_frames.append(
                self.scan._frame_flat_field_correction(
                    data=z_frame,
                    dark=dark,
                    flat_weights=(
                        self.scan._flats_weights[proj_i]
                        if proj_i in self.scan._flats_weights
                        else None
                    ),
                )
            )
        self._normed_volume = numpy.array(normed_frames)
        assert self._normed_volume.ndim == 3
        self._normed_sinogram_12 = self._normed_volume[:, 12, :]
        assert self._normed_sinogram_12.ndim == 2
        assert self._normed_sinogram_12.shape == (1503, 20)

    def testGetSinogram1(self):
        sinogram = self.scan.get_sinogram(line=12, subsampling=1)
        self.assertEqual(sinogram.shape, (1503, 20))

    def testGetSinogram2(self):
        """Test if subsampling is negative"""
        with self.assertRaises(ValueError):
            self.scan.get_sinogram(line=0, subsampling=-1)

    def testGetSinogram3(self):
        sinogram = self.scan.get_sinogram(line=0, subsampling=3)
        self.assertEqual(sinogram.shape, (501, 20))

    def testGetSinogram4(self):
        """Test if line is not in the projection"""
        with self.assertRaises(ValueError):
            self.scan.get_sinogram(line=-1, subsampling=1)

    def testGetSinogram5(self):
        """Test if line is not in the projection"""
        with self.assertRaises(ValueError):
            self.scan.get_sinogram(line=25, subsampling=1)

    def testGetSinogram6(self):
        """Test if line is not in the projection"""
        with self.assertRaises(TypeError):
            self.scan.get_sinogram(line=0, subsampling="tata")


class TestIgnoredProjections(HDF5TestBaseClass):
    """Test the ignore_projections parameter"""

    def setUp(self) -> None:
        super(TestIgnoredProjections, self).setUp()
        self.dataset_file = self.get_dataset("h5_datasets/frm_edftomomill_oneentry.nx")

    def testIgnoreProjectionsIndices(self):
        ignored_projs_indices = [387, 388, 389, 390, 391, 392, 393, 394, 395, 396]
        scan = NXtomoScan(
            scan=self.dataset_file,
            ignore_projections={"kind": "indices", "values": ignored_projs_indices},
        )
        for idx in ignored_projs_indices:
            self.assertFalse(
                idx in scan.projections,
                "Projection index %d is supposed to be ignored" % idx,
            )

    def testIgnoreProjectionsAngles(self):
        ignored_projs_angles = [102, 102.12, 102.24, 102.36, 103.8]
        corresponding_projs_indices = [872, 873, 874, 875, 887]  # mind the jump
        scan = NXtomoScan(
            scan=self.dataset_file,
            ignore_projections={"kind": "angles", "values": ignored_projs_angles},
        )
        self.assertTrue(
            numpy.allclose(
                corresponding_projs_indices, scan.get_ignored_projection_indices()
            )
        )
        for idx in scan.projections.keys():
            assert_func = (
                self.assertFalse
                if idx in corresponding_projs_indices
                else self.assertTrue
            )
            assert_func(
                idx in scan.projections,
                "Something off with projection index %d. The scan has projections %s, but should have the following excluded: %s"
                % (idx, sorted(scan.projections.keys()), corresponding_projs_indices),
            )

    def testIgnoreProjectionsAngularRange(self):
        ignored_projs_angular_range = [40.18, 58.8]
        corresponding_projs_indices = list(range(357, 512 + 1))
        scan = NXtomoScan(
            scan=self.dataset_file,
            ignore_projections={"kind": "range", "values": ignored_projs_angular_range},
        )
        self.assertTrue(
            numpy.allclose(
                corresponding_projs_indices, scan.get_ignored_projection_indices()
            )
        )
        for idx in scan.projections.keys():
            assert_func = (
                self.assertFalse
                if idx in corresponding_projs_indices
                else self.assertTrue
            )
            assert_func(
                idx in scan.projections,
                "Something off with projection index %d. The scan has projections %s, but should have the following excluded: %s"
                % (idx, sorted(scan.projections.keys()), corresponding_projs_indices),
            )


class TestGetSinogramLegacy(unittest.TestCase):
    def setUp(self) -> None:
        self.test_dir = tempfile.mkdtemp()

        self.proj_data = numpy.arange(1000, 1000 + 10 * 20 * 30).reshape(30, 10, 20)
        self.proj_angle = numpy.linspace(0, 180, 30)
        self.dark_value = 0.5
        self.flat_value = 1
        self.dark_data = numpy.ones((10, 20)) * self.dark_value
        self.dark_angle = 0
        self.flat_data_1 = numpy.ones((10, 20)) * self.flat_value
        self.flat_angle_1 = 0

        self.flat_data_2 = numpy.ones((10, 20)) * self.flat_value
        self.flat_angle_2 = 90
        self.flat_data_3 = numpy.ones((10, 20)) * self.flat_value
        self.flat_angle_3 = 180

        # data dataset
        self.data = numpy.empty((34, 10, 20))
        self.data[0] = self.dark_data
        self.data[1] = self.flat_data_1
        self.data[2:17] = self.proj_data[:15]
        self.data[17] = self.flat_data_2
        self.data[18:33] = self.proj_data[15:]
        self.data[33] = self.flat_data_3

        self.file_path = os.path.join(self.test_dir, "test.h5")
        self.create_arange_dataset(self.file_path)

    def create_arange_dataset(self, file_path):
        if os.path.exists(file_path):
            os.remove(file_path)

        with h5py.File(file_path, mode="a") as h5f:
            entry = h5f.require_group("entry0000")

            # rotation angle
            entry["instrument/detector/data"] = self.data
            rotation_angle = numpy.empty(34)
            rotation_angle[0] = self.dark_angle
            rotation_angle[1] = self.flat_angle_1
            rotation_angle[2:17] = self.proj_angle[:15]
            rotation_angle[17] = self.flat_angle_2
            rotation_angle[18:33] = self.proj_angle[15:]
            rotation_angle[33] = self.flat_angle_3

            entry["sample/rotation_angle"] = rotation_angle

            # image key / images keys
            image_keys = []
            image_keys.append(ImageKey.DARK_FIELD.value)
            image_keys.append(ImageKey.FLAT_FIELD.value)
            image_keys.extend([ImageKey.PROJECTION.value] * 15)
            image_keys.append(ImageKey.FLAT_FIELD.value)
            image_keys.extend([ImageKey.PROJECTION.value] * 15)
            image_keys.append(ImageKey.FLAT_FIELD.value)
            entry["instrument/detector/image_key"] = numpy.array(image_keys)
            entry["instrument/detector/image_key_control"] = numpy.array(image_keys)

    def tearDown(self) -> None:
        shutil.rmtree(self.test_dir)

    def testImplementations(self):
        scan = NXtomoScan(self.file_path, "entry0000")
        assert len(scan.projections) == 30
        assert len(scan.flats) == 3
        assert len(scan.darks) == 1

        scan.set_reduced_darks(
            {
                0: self.dark_data,
            }
        )

        scan.set_reduced_flats(
            {
                1: self.flat_data_1,
                17: self.flat_data_2,
                33: self.flat_data_3,
            }
        )

        scan._flats_weights = scan._get_flats_weights()
        sinogram_old = scan._get_sinogram_ref_imp(line=5)
        sinogram_new = scan.get_sinogram(line=5)
        raw_sinogram = self.proj_data[:, 5, :]
        corrected = (raw_sinogram - self.dark_value) / (
            self.flat_value - self.dark_value
        )
        numpy.testing.assert_array_equal(corrected, sinogram_new)
        numpy.testing.assert_array_equal(sinogram_old, sinogram_new)


def test_NXtomoScan_API():
    """several minor test of the NXtomoScan API"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        scan.clear_cache()
        scan.is_abort()
        scan.nexus_version = 1.1
        scan.sequence_name
        scan.sample_name
        scan.group_size
        scan.exposure_time
        with pytest.raises(NotImplementedError):
            scan.ff_interval


def test_NXtomoScan_source_API():
    """test dedicated API for Source"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
        field_of_view="Full",
    ) as scan:
        scan.source
        scan.source_name
        scan.source_type
        scan.instrument_name
        assert scan.field_of_view.value == "Full"
        scan.sample_x_pixel_size
        scan.sample_y_pixel_size


def test_TomoFrame_API():
    """Test TomoFrame API"""
    frame = TomoFrame(index=0, sequence_number=0)
    frame.image_key = ImageKey.PROJECTION
    with pytest.raises(TypeError):
        frame.image_key = "projection"
    frame.rotation_angle = 12.0
    frame.x_translation
    frame.y_translation
    frame.z_translation
    frame.is_control


def test_NXtomoScan_nxversion():
    """test various NX versions"""
    with NXtomoMockContext(
        scan_path=os.path.join(tempfile.mkdtemp(), "scan_test"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        hdf5scan = NXtomoScan(scan.master_file)
        # The default behavior is to take NX 1.1 if not present in file metadata
        # This is not future-proof, perhaps do a _NEXUS_PATHS_LATEST class ?
        assert hdf5scan.nexus_path is nx_tomo_path_latest
        energy_path_11 = hdf5scan.nexus_path.ENERGY_PATH

        hdf5scan = NXtomoScan(scan.master_file, nx_version=1.0)
        # Parse with a different NX version.
        # This Mock will provide energy at two different locations (instrument/beam/energy and beam/energy)
        # so we cannot test directly the "legacy files" behavior.
        # Instead we just check that the energy path is different in NX 1.0 and NX 1.1.
        assert hdf5scan.nexus_path is nx_tomo_path_v_1_0
        energy_path_10 = hdf5scan.nexus_path.ENERGY_PATH
        assert energy_path_10 != energy_path_11


def test_get_relative_file(tmpdir):
    """Test that get_relative_file function is working correctly for HDFScan"""
    folder_path = os.path.join(tmpdir, "scan_test")
    with NXtomoMockContext(
        scan_path=folder_path,
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        expected_f1 = os.path.join(folder_path, "scan_test_nabu_processes.h5")
        assert (
            scan.get_relative_file("nabu_processes.h5", with_dataset_prefix=True)
            == expected_f1
        )

        expected_f2 = os.path.join(folder_path, "nabu_processes.h5")
        assert (
            scan.get_relative_file("nabu_processes.h5", with_dataset_prefix=False)
            == expected_f2
        )


def test_save_and_load_dark(tmp_path):
    """test saving and loading of the dark is workinf for HDF5"""
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    with NXtomoMockContext(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        # 1. simple test on loading / saving the darks and flats
        assert scan.load_reduced_darks() == {}
        assert scan.load_reduced_flats() == {}
        dark_frame = numpy.ones((100, 100))
        scan.save_reduced_darks(
            {
                0: dark_frame,
            }
        )
        assert scan.load_reduced_flats() == {}
        loaded_darks = scan.load_reduced_darks()
        assert len(loaded_darks) == 1
        assert 0 in loaded_darks
        numpy.testing.assert_array_equal(loaded_darks[0], dark_frame)

        # 2. test overwrite and make sure we are properly removing the existing dataset
        output_file_name = (
            scan.REDUCED_DARKS_DATAURLS[0]
            .file_path()
            .format(
                entry=scan.entry,
                idx=0,
                idx_zfill4=str(0).zfill(4),
                scan_prefix=os.path.basename(
                    scan.path,
                ),
            )
        )
        output_file_path = os.path.join(scan.path, output_file_name)

        # 2.1 make sure when should overwrite the dataset the adding size is limited (and dataset is overwrite)
        first_file_size = os.path.getsize(output_file_path)
        scan.save_reduced_darks(
            {
                0: dark_frame,
            },
            overwrite=True,
        )
        assert os.path.getsize(output_file_path) < (
            first_file_size + dark_frame.nbytes / 2
        )
        # 2.2 test when the dataset cannot be overwrite
        scan.save_reduced_darks(
            {
                100: dark_frame,
            },
            overwrite=True,
        )
        loaded_darks = scan.load_reduced_darks()
        assert tuple(loaded_darks.keys()) == (100,)
        assert os.path.getsize(output_file_path) > (first_file_size + dark_frame.nbytes)


def test_save_and_load_flats(tmp_path):
    """test saving and loading of the flats is workinf for HDF5"""
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    with NXtomoMockContext(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        assert scan.load_reduced_darks() == {}
        assert scan.load_reduced_flats() == {}

        flat_frame_1 = numpy.ones((100, 100))
        flat_frame_222 = numpy.zeros((100, 100))
        scan.save_reduced_flats(
            {
                1: flat_frame_1,
                222: flat_frame_222,
            },
        )
        loaded_flats = scan.load_reduced_flats()
        assert len(loaded_flats) == 2
        assert 1 in loaded_flats
        assert 222 in loaded_flats
        numpy.testing.assert_array_equal(loaded_flats[1], flat_frame_1)
        numpy.testing.assert_array_equal(loaded_flats[222], flat_frame_222)
        assert scan.load_reduced_darks() == {}

        # test to save other flats to insure older one are removed
        flat_frame_333 = numpy.ones((100, 100)) * 3.2
        flat_frame_1 = numpy.ones((100, 100)) * 2.1
        scan.save_reduced_flats(
            {
                1: flat_frame_1,
                333: flat_frame_333,
            },
            overwrite=True,
        )
        loaded_flats = scan.load_reduced_flats()
        assert len(loaded_flats) == 2
        assert 1 in loaded_flats
        assert 333 in loaded_flats
        numpy.testing.assert_array_equal(loaded_flats[1], flat_frame_1)
        numpy.testing.assert_array_equal(loaded_flats[333], flat_frame_333)


def test_save_dark_flat_reduced_several_urls(tmp_path):
    """test saving and loading dark and flat providing several urls"""
    test_dir = tmp_path / "test_dir"
    os.makedirs(test_dir)
    with NXtomoMockContext(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        assert scan.load_reduced_darks() == {}
        assert scan.load_reduced_flats() == {}

        url_flats_edf = NXtomoScan.REDUCED_FLATS_DATAURLS[0]
        url_flats_processes = DataUrl(
            file_path=scan.get_relative_file(
                file_name="my_processes.h5", with_dataset_prefix=False
            ),
            data_path="/{entry}/process_1/results/flats/{index}",
            scheme="hdf5",
        )
        url_darks_edf = NXtomoScan.REDUCED_DARKS_DATAURLS[0]
        url_darks_processes = DataUrl(
            file_path=scan.get_relative_file(
                file_name="my_processes.h5", with_dataset_prefix=False
            ),
            data_path="/{entry}/process_1/results/darks/{index}",
            scheme="silx",
        )

        flat_frame_1 = numpy.ones((100, 100))
        flat_frame_222 = numpy.zeros((100, 100))
        flats_infos = ReducedFramesInfos()

        flats_infos.count_time = [
            1,
        ]
        flats_infos.machine_current = [
            12.3,
        ]

        with pytest.raises(ValueError):
            scan.save_reduced_flats(
                flats={
                    1: flat_frame_1,
                    222: flat_frame_222,
                },
                flats_infos=flats_infos,
            )

        flats_infos.count_time = [1.0, 1.0]
        flats_infos.machine_current = [12.3, 13.2]
        scan.save_reduced_flats(
            flats={
                1: flat_frame_1,
                222: flat_frame_222,
            },
            flats_infos=flats_infos,
        )

        assert scan.load_reduced_darks() == {}
        flat_frame_1 = numpy.ones((100, 100))
        flat_frame_1000 = numpy.ones((100, 100)) * 1.2
        dark_frame = numpy.zeros((100, 100))
        scan.save_reduced_flats(
            flats={
                1: flat_frame_1,
                1000: flat_frame_1000,
            },
            flats_infos=flats_infos,
            output_urls=(
                url_flats_edf,
                url_flats_processes,
            ),
            overwrite=True,
        )
        # test raise a type error if frame is not a numpy array
        with pytest.raises(TypeError):
            scan.save_reduced_darks(
                darks={
                    0: "test",
                },
                output_urls=(
                    url_darks_edf,
                    url_darks_processes,
                ),
                overwrite=True,
            )

        # test raise a type error if frame is not a 2D numpy array
        with pytest.raises(ValueError):
            scan.save_reduced_darks(
                darks={
                    0: numpy.asarray([12, 13]),
                },
                output_urls=(
                    url_darks_edf,
                    url_darks_processes,
                ),
                overwrite=True,
            )

        scan.save_reduced_darks(
            darks={
                0: dark_frame,
            },
            output_urls=(
                url_darks_edf,
                url_darks_processes,
            ),
            overwrite=True,
        )
        darks_infos = ReducedFramesInfos()
        darks_infos.count_time = [2.5]
        darks_infos.machine_current = [13.1]
        scan.save_reduced_darks(
            darks={
                0: dark_frame,
            },
            output_urls=(
                url_darks_edf,
                url_darks_processes,
            ),
            darks_infos=darks_infos,
            overwrite=True,
        )
        assert len(scan.load_reduced_flats(return_info=False)) == 2
        assert len(scan.load_reduced_darks(return_info=False)) == 1

        _, loaded_darks_infos = scan.load_reduced_darks(return_info=True)
        assert loaded_darks_infos == darks_infos

        processes_file = os.path.join(scan.path, "my_processes.h5")
        assert os.path.exists(processes_file)

        _, loaded_flats_infos = scan.load_reduced_flats(return_info=True)
        assert loaded_flats_infos == flats_infos

        with h5py.File(processes_file, mode="r") as h5s:
            assert "/entry/process_1/results/darks" in h5s
            darks_grp = h5s["/entry/process_1/results/darks"]
            assert "0" in darks_grp
            numpy.testing.assert_array_equal(darks_grp["0"][()], dark_frame)
            assert "/entry/process_1/results/flats" in h5s
            flats_grp = h5s["/entry/process_1/results/flats"]
            assert "1" in flats_grp
            assert "1000" in flats_grp
            numpy.testing.assert_array_equal(flats_grp["1"], flat_frame_1)
            numpy.testing.assert_array_equal(flats_grp["1000"], flat_frame_1000)

        assert len(scan.load_reduced_flats((url_flats_processes,))) == 2
        assert len(scan.load_reduced_darks((url_darks_processes,))) == 1
        loaded_reduced_darks = scan.load_reduced_darks(
            (url_darks_processes,), return_as_url=True
        )
        assert isinstance(loaded_reduced_darks[0], DataUrl)
        assert loaded_reduced_darks[0].file_path() == processes_file
        assert loaded_reduced_darks[0].data_path() == "/entry/process_1/results/darks/0"
        assert loaded_reduced_darks[0].scheme() == "silx"


def test_get_bounding_box(tmp_path):
    """
    check bounding box and overlap works.
    Warning: the MockNxtomo is using the NXtomo coordinate system ()

                          Z axis
                             ^    Y axis
                             | /
           x-ray             |/
           -------->          ------> X axis

    When tomoscan coordinate system is the ESRF one

                            Y axis
                              ^   X axis
                              |  /
            x-ray             | /
            -------->          ------> Z axis

    """
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    x_pos_nxtomo_coord = 0.0 * _ureg.millimeter
    y_pos_nxtomo_coord = 1.0 * _ureg.millimeter
    z_pos_nxtomo_coord = 3.6 * _ureg.millimeter
    magnification = 2

    scan = MockNXtomo(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
        magnification=magnification,
        x_pos=x_pos_nxtomo_coord.magnitude,
        y_pos=y_pos_nxtomo_coord.magnitude,
        z_pos=z_pos_nxtomo_coord.magnitude,
    ).scan

    pixel_size = MockNXtomo.SAMPLE_PIXEL_SIZE * _ureg.micrometer
    frame_dim = 200

    # note: pixel size is provided with magnificaton. To get
    min_x_nxtomo_coord = x_pos_nxtomo_coord.to(_ureg.meter).magnitude - (
        frame_dim / 2.0 * pixel_size.to(_ureg.meter).magnitude
    )
    max_x_nxtomo_coord = x_pos_nxtomo_coord.to(_ureg.meter).magnitude + (
        frame_dim / 2.0 * pixel_size.to(_ureg.meter).magnitude
    )
    assert scan.get_bounding_box(axis="y") == (min_x_nxtomo_coord, max_x_nxtomo_coord)

    min_y_nxtomo_coord = y_pos_nxtomo_coord.to(_ureg.meter).magnitude - (
        frame_dim / 2.0 * pixel_size.to(_ureg.meter).magnitude
    )
    max_y_nxtomo_coord = y_pos_nxtomo_coord.to(_ureg.meter).magnitude + (
        frame_dim / 2.0 * pixel_size.to(_ureg.meter).magnitude
    )
    assert scan.get_bounding_box(axis="z") == (min_y_nxtomo_coord, max_y_nxtomo_coord)

    min_z_nxtomo_coord = z_pos_nxtomo_coord.to(_ureg.meter).magnitude - (
        frame_dim / 2.0 * pixel_size.to(_ureg.meter).magnitude
    )
    max_z_nxtomo_coord = z_pos_nxtomo_coord.to(_ureg.meter).magnitude + (
        frame_dim / 2.0 * pixel_size.to(_ureg.meter).magnitude
    )
    assert scan.get_bounding_box(axis="x") == (min_z_nxtomo_coord, max_z_nxtomo_coord)
    assert scan.get_bounding_box(axis=None) is not None


def test_clearing_cache(tmp_path):
    with NXtomoMockContext(
        scan_path=str(tmp_path / "my_scan"),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        """Minor test on clearing cache."""
        assert scan.darks not in (None, {})
        assert scan.flats not in (None, {})
        nexus_path = get_nexus_paths(None)
        with h5py.File(scan.master_file, mode="a") as h5f:
            del h5f[f"{scan.entry}/{nexus_path.IMG_KEY_PATH}"]
            h5f[f"{scan.entry}/{nexus_path.IMG_KEY_PATH}"] = [
                ImageKey.INVALID.value
            ] * len(scan._image_keys_control)
            del h5f[f"{scan.entry}/{nexus_path.IMG_KEY_CONTROL_PATH}"]
            h5f[f"{scan.entry}/{nexus_path.IMG_KEY_CONTROL_PATH}"] = [
                ImageKey.INVALID.value
            ] * len(scan._image_keys_control)
        assert scan.darks not in (None, {})
        scan.clear_cache()
        assert scan.darks in (None, {})
        assert scan.flats in (None, {})


configs = (
    {
        "n_proj": 10,
        "with_darks": True,
        "with_flats": True,
        "return_angles": (0,),
    },
    {
        "n_proj": 10,
        "with_darks": False,
        "with_flats": False,
        "return_angles": tuple(),
    },
)


@pytest.mark.parametrize("config", configs)
def test_nxtomo_without_key_control(tmp_path, config):
    scan_path = str(tmp_path / "scan_no_image_key_control")
    mock = MockNXtomo(
        scan_path=scan_path,
        n_proj=config["n_proj"],
        n_ini_proj=config["n_proj"],
        n_alignement_proj=0,
        create_ini_dark=config["with_darks"],
        create_ini_flat=config["with_flats"],
        create_final_flat=False,
    )

    for i_frame, angle in enumerate(config["return_angles"]):
        mock.add_alignment_radio(index=i_frame + config["n_proj"], angle=angle)

    scan = mock.scan
    with h5py.File(scan.master_file, mode="a") as h5f:
        entry = h5f[scan.entry]
        del entry["instrument/detector/image_key_control"]

    clean_scan = NXtomoScan(scan=scan.master_file, entry=scan.entry)
    assert len(clean_scan.projections) == config["n_proj"] + len(
        config["return_angles"]
    )
    # whrn there is not 'image_key_control' we consider there is no alignment / return projections
    assert len(clean_scan.return_projs) == 0
    assert (len(clean_scan.darks) > 0) == config["with_darks"]
    assert (len(clean_scan.flats) > 0) == config["with_flats"]


def test_build_drac_metadata(tmp_path):
    scan_path = str(tmp_path / "test_build_drac_metadata.hdf5")
    scan = MockNXtomo(
        scan_path=scan_path,
        n_proj=10,
        n_ini_proj=10,
        n_alignement_proj=0,
        create_ini_dark=True,
        create_ini_flat=True,
        create_final_flat=True,
    ).scan
    # append some metadata to the file
    nx_tomo = NXtomo().load(
        scan.master_file, "entry", detector_data_as="as_numpy_array"
    )
    # warning: the Y on the NXtomo side match the MCStats convention when tomoscan coordinate system is the one used at ESRF.
    nx_tomo.sample.y_translation = (
        numpy.concatenate(
            (
                [-1],  # dark frame
                [0] * 10,  # flat frames
                numpy.linspace(1, 1.1, num=10, endpoint=False),  # projections
                [2] * 10,  # flat
            )
        )
        * _ureg.meter
    )
    nx_tomo.sample.x_pixel_size = 2.3 * _ureg.micrometer
    nx_tomo.instrument.detector.distance = 5.6 * _ureg.micrometer

    nx_tomo.save(scan.master_file, "entry", overwrite=True)

    icat_metadata = scan.build_drac_metadata()

    assert numpy.isclose(icat_metadata.pop("TOMO_ZDelta"), 0.01)
    assert numpy.isclose(icat_metadata.pop("TOMO_detectorDistance"), 5.6)
    assert numpy.isclose(icat_metadata.pop("TOMO_pixelSize"), 2.3)
    assert icat_metadata == {
        "definition": "TOMO",
        "TOMO_accExposureTime": numpy.float64(1.0),
        "TOMO_darkN": 1,
        "definition": "TOMO",
        "TOMO_halfAcquisition": False,
        "TOMO_projN": 10,
        "TOMO_refN": 10,
        "sample_name": "test",
        "Sample_name": "test",
        "TOMO_scanRange": 360,
        "TOMO_start_angle": numpy.float64(0.0),
        "TOMO_x_pixel_n": 200,
        "TOMO_y_pixel_n": 200,
        "TOMO_ZStart": 1.0,
    }


def test_splitted_flat_series(tmp_path):
    """test splitted_flat_serie property"""

    file_path = str(tmp_path / "test_flat_split.nx")

    nx_tomo_writer = NXtomo()
    nx_tomo_reader = NXtomoScan(scan=file_path, entry="entry")

    def setImageKey(values):
        nx_tomo_writer.sample.rotation_angle = numpy.linspace(0, 360, 10) * _ureg.degree
        nx_tomo_writer.instrument.detector.data = numpy.arange(10).reshape(10, 1, 1)
        nx_tomo_writer.instrument.detector.image_key_control = values
        nx_tomo_writer.save(file_path, "entry", overwrite=True)
        nx_tomo_reader.clear_frames_cache()

    setImageKey([ImageKey.FLAT_FIELD.value] * 10)

    assert len(nx_tomo_reader.splitted_flat_series) == 1
    assert len(nx_tomo_reader.splitted_flat_series[0]) == 10

    setImageKey(
        numpy.concatenate(
            (
                [ImageKey.FLAT_FIELD.value] * 3,
                [ImageKey.PROJECTION.value] * 3,
                [ImageKey.FLAT_FIELD.value] * 4,
            )
        )
    )

    assert len(nx_tomo_reader.splitted_flat_series) == 2
    assert len(nx_tomo_reader.splitted_flat_series[0]) == 3
    assert len(nx_tomo_reader.splitted_flat_series[1]) == 4

    setImageKey([ImageKey.PROJECTION.value] * 10)
    nx_tomo_reader._image_keys = nx_tomo_reader._image_keys_control = [
        ImageKey.PROJECTION.value
    ] * 10
    assert len(nx_tomo_reader.splitted_flat_series) == 0


def test_get_propagation_distance(tmp_path):
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.distance = 0.3 * _ureg.cm
    nx_tomo.instrument.source.distance = 1.2 * _ureg.m
    nx_tomo.sample.propagation_distance = (0.3 * _ureg.cm * 1.2 * _ureg.m) / (
        1.2 * _ureg.m + 0.3 * _ureg.cm
    )

    output_file = tmp_path / "my_nxtomo.nx"
    entry_name = "entry"

    nx_tomo.save(str(output_file), entry_name)

    scan = NXtomoScan(output_file, entry_name)
    assert (
        scan.sample_detector_distance
        == nx_tomo.instrument.detector.distance.to_base_units().magnitude
    )
    assert (
        scan.source_sample_distance
        == nx_tomo.instrument.source.distance.to_base_units().magnitude
    )
    z1 = nx_tomo.instrument.source.distance
    z2 = nx_tomo.instrument.detector.distance
    assert (
        scan.propagation_distance == ((z1 * z2) / (z1 + z2)).to_base_units().magnitude
    )


def test_pixel_size_getter(tmp_path):
    nx_tomo = NXtomo()
    nx_tomo.instrument.detector.x_pixel_size = 0.03 * _ureg.micrometer
    nx_tomo.instrument.detector.y_pixel_size = 0.02 * _ureg.micrometer

    nx_tomo.sample.x_pixel_size = 0.1 * _ureg.micrometer
    nx_tomo.sample.y_pixel_size = 0.12 * _ureg.micrometer

    output_file = tmp_path / "my_nxtomo.nx"
    entry_name = "entry"

    nx_tomo.save(str(output_file), entry_name)

    scan = NXtomoScan(output_file, entry_name)

    assert (
        scan.detector_x_pixel_size
        == nx_tomo.instrument.detector.x_pixel_size.to_base_units().magnitude
    )
    assert (
        scan.detector_y_pixel_size
        == nx_tomo.instrument.detector.y_pixel_size.to_base_units().magnitude
    )

    assert (
        scan.sample_x_pixel_size
        == nx_tomo.sample.x_pixel_size.to_base_units().magnitude
    )
    assert (
        scan.sample_y_pixel_size
        == nx_tomo.sample.y_pixel_size.to_base_units().magnitude
    )



from __future__ import annotations

import os
import numpy
import h5py

from silx.io.url import DataUrl

from tomoscan.framereducer.ReduceFrameSaver import ReduceFrameSaver
from tomoscan.framereducer.reducedframesinfos import ReducedFramesInfos


def test_ReduceFrameSaver(tmp_path):
    """
    Make sure the API of 'ReducedFramesInfos' is accessible as a standalone
    """
    output_file = os.path.join(tmp_path, "output.hdf5")
    frames_metadata = ReducedFramesInfos()
    frames_metadata.machine_current = (12.3,)
    frames_metadata.count_time = (1.0,)

    ReduceFrameSaver(
        frames={
            0: numpy.ones((100, 100)),
        },
        output_urls=(
            DataUrl(
                file_path=output_file,
                data_path="entry0000/flats/{index}",
                scheme="silx",
            ),
        ),
        metadata_output_urls=(
            DataUrl(
                file_path=output_file,
                data_path="entry0000/flats/",
                scheme="silx",
            ),
        ),
        frames_metadata=frames_metadata,
    ).save()

    assert os.path.exists(output_file)
    with h5py.File(output_file, mode="r") as h5f:
        my_flat_dataset = h5f["entry0000/flats/0"]
        numpy.testing.assert_almost_equal(my_flat_dataset[:], numpy.ones((100, 100)))

        count_time = h5f["entry0000/flats/count_time"]
        numpy.testing.assert_almost_equal(
            count_time[:],
            numpy.array(
                [
                    1,
                ]
            ),
        )
        machine_current = h5f["entry0000/flats/machine_current"]
        numpy.testing.assert_almost_equal(
            machine_current[:],
            numpy.array(
                [
                    12.3,
                ]
            ),
        )



"""mapper functions to convert from raw data fields (like nabu reconstructed volume metadata) to drac / icat fields"""

from __future__ import annotations

import numpy
from .recons_vol_mapping import DRAC_NABU_MAPPING, DracRawDataMapping


__all__ = [
    "flatten_vol_metadata",
    "map_nabu_keys_to_drac",
]


def flatten_vol_metadata(metadata: dict, parent_key: str = "", sep="/") -> dict:
    """
    Flatten a nested dictionary (expected to come from nabu volume metadata) and converting arrays to lists
    and numpy types to serializable types. The resulting dictionary is JSON-serializable.

    :param metadata: The dictionary to flatten.
    :param parent_key: The base key string used for recursive calls.
    :param sep: The separator used between parent and child keys.

    :return: The flattened and type-converted dictionary.
    """
    # Exclusion list
    items = []
    for k, v in metadata.items():

        # Construct the new key
        new_key = f"{parent_key}{sep}{k}" if parent_key else k

        if isinstance(v, dict):
            items.extend(flatten_vol_metadata(v, new_key, sep=sep).items())
        else:
            # Convert numpy arrays to lists
            if isinstance(v, numpy.ndarray):
                v = v.tolist()
            # Convert numpy boolean to native boolean
            elif isinstance(v, numpy.bool_):
                v = bool(v)
            # Convert numpy integers to native int
            elif isinstance(v, (numpy.int64, numpy.int32)):
                v = int(v)
            # Convert numpy floats to native float
            elif isinstance(v, (numpy.float64, numpy.float32)):
                v = float(v)

            items.append((new_key, v))

    return dict(items)


def map_nabu_keys_to_drac(flatten_dict: dict) -> dict:
    """
    map a set of **nabu** keys to drac 'TOMOReconstruction' definition (https://gitlab.esrf.fr/icat/hdf5-master-config/-/blob/master/hdf5_cfg.xml)
    """
    drac_dict = {}
    for map in DRAC_NABU_MAPPING:
        assert isinstance(
            map, DracRawDataMapping
        ), f"map should be an instance of {DracRawDataMapping}. Got {type(map)}"
        if map.raw_data_key in flatten_dict:
            raw_data = flatten_dict[map.raw_data_key]
            drac_value = map.to_drac(raw_data)
            drac_dict[map.drac_key] = drac_value

    return drac_dict



# coding: utf-8
"""module defining utils for a jp2k volume"""
from __future__ import annotations

from __future__ import annotations

import logging
import os
import re
from warnings import warn

import numpy
from silx.io.dictdump import dicttoini
from silx.io.dictdump import load as load_ini
from silx.io.url import DataUrl

from tomoscan.scanbase import TomoScanBase
from tomoscan.utils import docstring
from tomoscan.volumebase import VolumeBase, SliceTuple
from tomoscan.io import cast_to_default_types
from tomoscan.utils.io import (
    deprecated_warning,
    _deprecate_url_in_signature,
    catch_log_messages,
)

_logger = logging.getLogger(__name__)


__all__ = ["VolumeSingleFrameBase", "TIFFLikeDiskAccessor"]


class VolumeSingleFrameBase(VolumeBase):
    """
    Base class for Volume where each slice is saved in a separate file like edf, jp2k or tiff.

    :param start_index: users can provide a shift on fill name when saving the file. This is interesting if you want to create
                        create a volume from several writer.
    """

    DEFAULT_DATA_SCHEME = None

    DEFAULT_DATA_PATH_PATTERN = "{volume_basename}_{index_zfill6}.{data_extension}"

    DEFAULT_METADATA_EXTENSION = "txt"

    # information regarding metadata
    DEFAULT_METADATA_SCHEME = "ini"

    DEFAULT_METADATA_PATH_PATTERN = "{volume_basename}_infos.{metadata_extension}"

    def __init__(
        self,
        url: DataUrl | None = None,
        data: numpy.ndarray | None = None,
        source_scan: TomoScanBase | None = None,
        metadata: dict | None = None,
        data_url: DataUrl | None = None,
        metadata_url: DataUrl | None = None,
        overwrite: bool = False,
        start_index: int = 0,
        volume_basename: str | None = None,
        data_extension=None,
        metadata_extension="txt",
    ) -> None:
        self._volume_basename = volume_basename
        super().__init__(
            url,
            data,
            source_scan,
            metadata,
            data_url,
            metadata_url,
            overwrite,
            data_extension,
            metadata_extension,
        )
        self._start_index = start_index
        self.write_in_descending_order = False
        self._skip_existing_data_files_removal = False

    def get_first_file(self):
        try:
            first_file = next(self.browse_data_files())
        except StopIteration:
            first_file = None
        return first_file

    def get_file_slice_index(self, filename):
        if (
            self.DEFAULT_DATA_PATH_PATTERN
            != VolumeSingleFrameBase.DEFAULT_DATA_PATH_PATTERN
        ):
            raise ValueError("Non-default file name pattern not handled")
        return int(filename.split(".")[0].split("_")[-1])

    @property
    def start_index(self) -> int:
        return self._start_index

    @start_index.setter
    def start_index(self, start_idx: int) -> None:
        self._start_index = start_idx

    def get_volume_basename(self, url=None):
        if self._volume_basename is not None:
            return self._volume_basename
        else:
            url = url or self.data_url
            return os.path.basename(url.file_path())

    @property
    def skip_existing_data_files_removal(self) -> bool:
        """The loading of the volume for single frame base is done by loading all the file
        contained in a folder data_url.file_path(). When saving the data we make sure there is no 'remaining' of
        any previous saving by using the file pattern.
        But when we want to save a volume from several thread (one thread save the n first frame, second the n next frame ...) this
        could be a limitation.
        So in this case we can use the 'ignore_existing_files' that will avoid calling '_remove_existing_data_files'
        """
        return self._skip_existing_data_files_removal

    @skip_existing_data_files_removal.setter
    def skip_existing_data_files_removal(self, ignore: bool) -> None:
        if not isinstance(ignore, bool):
            raise TypeError(f"ignore should be a bool. Got {type(ignore)}")
        self._skip_existing_data_files_removal = ignore

    @docstring(VolumeBase)
    def deduce_data_and_metadata_urls(self, url: DataUrl | None) -> tuple:
        """
        Deduce automatically data and metadata url.
        Default data will be saved as single frame edf.
        Default metadata will be saved as a text file
        """
        if url is None:
            return None, None
        else:
            metadata_keywords = {
                "volume_basename": self.get_volume_basename(url),
                "metadata_extension": self.metadata_extension,
            }
            metadata_data_path = self.DEFAULT_METADATA_PATH_PATTERN.format(
                **metadata_keywords
            )

            return (
                # data url
                DataUrl(
                    file_path=url.file_path(),
                    data_path=self.DEFAULT_DATA_PATH_PATTERN,
                    scheme=url.scheme() or self.DEFAULT_DATA_SCHEME,
                    data_slice=url.data_slice(),
                ),
                # medata url
                DataUrl(
                    file_path=url.file_path(),
                    data_path=metadata_data_path,
                    scheme=url.scheme() or self.DEFAULT_METADATA_SCHEME,
                ),
            )

    @docstring(VolumeBase)
    def load_metadata(self, url: DataUrl | None = None, store: bool = True) -> dict:
        _deprecate_url_in_signature(url=url)
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        if url.scheme() == "ini":
            metadata_file = url.file_path()
            if url.data_path() is not None:
                metadata_file = os.path.join(metadata_file, url.data_path())
                _logger.info(f"load data to {metadata_file}")
            try:
                metadata = load_ini(metadata_file, "ini")
            except FileNotFoundError:
                _logger.warning(
                    f"unable to load metadata from {metadata_file} - File not found"
                )
                metadata = {}
            except Exception as e:
                _logger.error(
                    f"Failed to load metadata from {metadata_file}. Error is {e}"
                )
                metadata = {}
        else:
            raise ValueError(f"scheme {url.scheme()} is not handled")

        if store:
            self.metadata = metadata
        return metadata

    @docstring(VolumeBase)
    def save_metadata(self, url: DataUrl | None = None) -> None:
        _deprecate_url_in_signature(url=url)
        if self.metadata is None:
            raise ValueError("No data to be saved")
        url = url or self.metadata_url
        if url is None:
            raise ValueError(
                "Cannot get metadata_url. An url should be provided. Don't know where to save this."
            )
        else:
            if url.scheme() == "ini":
                metadata_file = url.file_path()
                if url.data_path() is not None:
                    metadata_file = os.path.join(metadata_file, url.data_path())
                    _logger.info(f"save data to {metadata_file}")
                    if len(self.metadata) > 0:
                        metadata = cast_to_default_types(self.metadata)
                        dicttoini(metadata, metadata_file)
            else:
                raise ValueError(f"scheme {url.scheme()} is not handled")

    # utils to format file path

    def format_data_path_for_data(
        self, data_path: str, index: int, volume_basename: str
    ) -> str:
        """
        Return file path to save the frame at `index` of the current volume
        """
        keywords = {
            "index_zfill4": str(index + self.start_index).zfill(4),
            "index_zfill6": str(index + self.start_index).zfill(6),
            "volume_basename": volume_basename,
            "data_extension": self.data_extension,
        }
        return data_path.format(**keywords)

    def get_data_path_pattern_for_data(
        self, data_path: str, volume_basename: str
    ) -> str:
        """
        Return file path **pattern** (and not full path) to  load data.
        For example in edf it can return 'myacquisition_*.edf' in order to be handled by
        """
        keywords = {
            "index_zfill4": "[0-9]{3,4}",
            "index_zfill6": "[0-9]{3,6}",
            "volume_basename": volume_basename,
            "data_extension": self.data_extension,
        }
        return data_path.format(**keywords)

    def remove_existing_data_files(self, url: DataUrl | None = None) -> None:
        """Clean any existing files (if overwrite and rights) that must be used for saving"""
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if not os.path.exists(url.file_path()):
            return

        research_pattern = self.get_data_path_pattern_for_data(
            data_path=url.data_path(),
            volume_basename=self.get_volume_basename(url=url),
        )
        try:
            research_pattern = re.compile(research_pattern)
        except Exception:
            _logger.error(
                f"Fail to compute regular expression for {research_pattern}. Unable to load data"
            )
            return None
        for file_ in sorted(os.listdir(url.file_path())):
            if research_pattern.match(file_):
                full_file_path = os.path.join(url.file_path(), file_)
                if self.overwrite:
                    os.remove(full_file_path)
                else:
                    raise OSError(
                        f"Output file '{full_file_path}' already exists. No overwrite required. Skip writting"
                    )

    @docstring(VolumeBase)
    def save_data(self, url: DataUrl | None = None) -> None:
        _deprecate_url_in_signature(url=url)
        if self.data is None:
            raise ValueError("No data to be saved")
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )
        else:
            if not self.skip_existing_data_files_removal:
                self.remove_existing_data_files(url=url)

            _logger.info(f"save data to {url.path()}")
            # if necessary create output directory (some third part writer does not do it for us)
            try:
                os.makedirs(url.file_path(), exist_ok=True)
            except FileNotFoundError:
                # can raise FileNotFoundError if file path is '.' for example
                pass

            assert self.data.ndim == 3
            for frame, frame_dumper in zip(
                self.data,
                self.data_file_saver_generator(
                    n_frames=self.data.shape[0], data_url=url, overwrite=self.overwrite
                ),
            ):
                frame_dumper[:] = frame

    def data_file_name_generator(self, n_frames, data_url):
        """
        browse output files for n_frames
        """
        indices = range(n_frames)
        if self.write_in_descending_order:
            indices = list(range(0, -n_frames, -1))
        for i_frame in indices:
            file_name = self.format_data_path_for_data(
                data_url.data_path(),
                index=i_frame,
                volume_basename=self.get_volume_basename(data_url),
            )
            file_name = os.path.join(data_url.file_path(), file_name)
            yield file_name

    @docstring(VolumeBase)
    def data_file_saver_generator(
        self, n_frames, data_url: DataUrl | None = None, overwrite: bool = False
    ):
        class _FrameDumper:
            def __init__(self, url_scheme, file_name, callback) -> None:
                self.url_scheme = url_scheme
                self.file_name = file_name
                self.overwrite = overwrite
                self.__callback = callback

            def __setitem__(self, key, value):
                if not self.overwrite and os.path.exists(self.file_name):
                    raise OSError(
                        f"{self.file_name} already exists. If you want you can ask for the volume to overwriting existing files."
                    )
                if key != slice(None, None, None):
                    raise ValueError("item setting only handle ':' for now")
                self.__callback(
                    frame=value, file_name=self.file_name, scheme=self.url_scheme
                )

        # TODO: deprecate `data_url` parameter in the future.
        data_url = data_url or self.data_url
        os.makedirs(data_url.file_path(), exist_ok=True)
        for file_name in self.data_file_name_generator(
            n_frames=n_frames, data_url=data_url
        ):
            yield _FrameDumper(
                file_name=file_name,
                url_scheme=data_url.scheme(),
                callback=self.save_frame,
            )

    def get_volume_shape(self, url=None):
        _deprecate_url_in_signature(url=url)
        if self.data is not None:
            return self.data.shape
        else:
            first_slice = next(self.browse_slices(url=url))
            n_slices = len(tuple(self.browse_data_urls()))
            return n_slices, first_slice.shape[0], first_slice.shape[1]

    @docstring(VolumeBase)
    def load_data(
        self, url: DataUrl | None = None, store: bool = True
    ) -> numpy.ndarray:
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is None:
            raise ValueError(
                "Cannot get data_url. An url should be provided. Don't know where to save this."
            )
        data = list(self.browse_slices(url=url))

        if data == []:
            data = None
            _logger.warning(
                f"Failed to load any data for {self.get_identifier().short_description}"
            )
        else:
            data = numpy.asarray(data)
            if data.ndim != 3:
                raise ValueError(f"data is expected to be 3D not {data.ndim}.")

        if store:
            self.data = data

        return data

    def _read_frame_from_fast_axis(self, index: int, url: DataUrl) -> numpy.ndarray:
        """
        Read a single frame along the 'fast axis' (0 aka Z).
        Raise "IndexError if not accessible
        """
        data_files = tuple(self.browse_data_files())
        file_path = data_files[index]
        return self.load_frame(file_name=file_path, scheme=url.scheme())

    def get_slice(
        self,
        index: int | str = None,
        axis=None,
        xy=None,
        xz=None,
        yz=None,
        url: DataUrl | None = None,
    ):
        _deprecate_url_in_signature(url=url)
        # if we are on the 'fast' read axis then we can directly read a single file. Else we must fall back on the default reading mechanism
        if axis == 0 or xy is not None:
            if xy is not None:
                deprecated_warning(
                    type_="parameter",
                    name="xy",
                    replacement="axis and index",
                )
                if axis is None and index is None:
                    axis = 0
                    index = xy
                else:
                    raise ValueError("several axis (previously xy, xz, yz requested")
            return self._read_frame_from_fast_axis(
                index=index, url=(url or self.data_url)
            )
        else:
            return super().get_slice(index=index, axis=axis, xy=xy, xz=xz, yz=yz)

    def save_frame(self, frame: numpy.ndarray, file_name: str, scheme: str):
        """
        Function dedicated for volune saving each frame on a single file

        :param frame: frame to be save
        :param file_name: path to store the data
        :param scheme: scheme to save the data
        """
        raise NotImplementedError("Base class")

    def load_frame(self, file_name: str, scheme: str) -> numpy.ndarray:
        """
        Function dedicated for volume saving each frame on a single file

        :param file_name: path to store the data
        :param scheme: scheme to save the data
        """
        raise NotImplementedError("Base class")

    @docstring(VolumeBase)
    def load_chunk(self, chunk, url=None):
        if (not numpy.iterable(chunk)) or len(chunk) != 3:
            raise ValueError(
                f"Expected 'chunk' to be in the form (slice(start0, stop0), slice(start1, end1), slice(start2, end2)) but got: {chunk}"
            )
        url = url or self.data_url
        scheme = url.scheme() or self.DEFAULT_DATA_SCHEME
        files = list(self.browse_data_files())
        files_selection = files[chunk[0]]
        if files_selection == []:
            warn(
                f"load_chunk(): empty files selection returned by chunk={chunk[0]}",
                RuntimeWarning,
            )
            return numpy.array([], dtype="f")

        # Pre-allocate chunk
        first_frame = self.load_frame(files_selection[0], scheme=scheme)
        first_frame = first_frame[chunk[1], chunk[2]]
        result_chunk_shape = (len(files_selection),) + first_frame.shape
        result_chunk = numpy.zeros(result_chunk_shape, dtype=first_frame.dtype)

        # Browse files, populate chunk
        for i, filename in enumerate(files_selection):
            data = self.load_frame(filename, scheme=scheme)
            result_chunk[i] = data[chunk[1], chunk[2]]
        return result_chunk

    @docstring(VolumeBase)
    def browse_metadata_files(self, url=None):
        _deprecate_url_in_signature(url=url)
        url = url or self.metadata_url
        if url is None:
            return
        elif url.file_path() is not None:
            if url.scheme() == "ini":
                metadata_file = url.file_path()
                if url.data_path() is not None:
                    metadata_file = os.path.join(metadata_file, url.data_path())
                    if os.path.exists(metadata_file):
                        yield metadata_file
            else:
                raise ValueError(f"scheme {url.scheme()} is not handled")

    @docstring(VolumeBase)
    def browse_data_files(self, url=None):
        _deprecate_url_in_signature(url=url)
        url = url or self.data_url
        if url is None:
            return
        research_pattern = self.get_data_path_pattern_for_data(
            url.data_path(), volume_basename=self.get_volume_basename(url)
        )
        try:
            research_pattern = re.compile(research_pattern)
        except Exception:
            _logger.error(
                f"Fail to compute regular expression for {research_pattern}. Unable to load data"
            )
            return None

        # use case of a single file
        if not os.path.exists(url.file_path()):
            return
        elif os.path.isfile(url.file_path()):
            yield url.file_path()
        else:
            for file_ in sorted(os.listdir(url.file_path())):
                if research_pattern.match(file_):
                    full_file_path = os.path.join(url.file_path(), file_)
                    yield full_file_path

    @docstring(VolumeBase)
    def browse_data_urls(self, url=None):
        _deprecate_url_in_signature(url=url)
        url_ = url or self.data_url
        with catch_log_messages():
            for data_file in self.browse_data_files(url=url):
                yield DataUrl(
                    file_path=data_file,
                    scheme=url_.scheme(),
                )

    @docstring(VolumeBase)
    def browse_slices(self, url=None):
        _deprecate_url_in_signature(url=url)
        if url is None and self.data is not None:
            for data_slice in self.data:
                yield data_slice
        else:
            url_ = url or self.data_url
            if url_ is None:
                raise ValueError(
                    "No data and data_url know and no url provided. Unable to browse slices"
                )

            with catch_log_messages():
                for file_path in self.browse_data_files(url=url):
                    yield self.load_frame(file_name=file_path, scheme=url_.scheme())

    @property
    def write_in_descending_order(self):
        """
        Return True if the data is saved with decreasing indices:
          - data[0] is written to file with index `start_index`
          - data[1] is written to file with index `start_index - 1`
          - and so on
        """
        return self._write_in_descending_order

    @write_in_descending_order.setter
    def write_in_descending_order(self, val):
        self._write_in_descending_order = bool(val)


class TIFFLikeDiskAccessor:
    """
    Class to define a common implementation of '_get_slices_from_disk' for 'tiff like' format (!!! single frame per file !!!).
    (format unable to implement 'read_n_lines_in_file', 'read_n_columns_in_file' and 'read_file' functions

    .. warning:: a class that want to inherit from TIFFLikeDiskAccessor should also inherit from VolumeSingleFrameBase.
        a separated class has been created instead of a direct inheritance to clearly separate the '_get_slices_from_disk' instantiation
    """

    def _get_slices_from_disk(
        self, slices: tuple[SliceTuple], url: DataUrl | None = None
    ) -> dict[SliceTuple, numpy.ndarray]:

        _logger.debug("Try to read %s from disk", slices)
        volume_shape = self.get_volume_shape()  # pylint: disable=E1101
        _logger.debug("volume shape is %s", volume_shape)

        def get_output_slice_shape(axis):
            if axis == 0:
                return (volume_shape[1], volume_shape[2])
            elif axis == 1:
                return (volume_shape[0], volume_shape[2])
            elif axis == 2:
                return (volume_shape[0], volume_shape[1])
            else:
                raise ValueError(f"unknown requested axis ({axis})")

        # init resulting data
        result = {
            slice_: numpy.empty(shape=get_output_slice_shape(slice_.axis))
            for slice_ in slices
        }
        _logger.debug(
            "Reserve numpy arrays for reading %s",
            {slice_: (data.shape, data.dtype) for (slice_, data) in result.items()},
        )
        from tomoscan.esrf.volume.utils import (
            group_slices_by_axis,
        )  # avoid cyclic import

        indices_to_read_per_axis: dict[int, set[int]] = group_slices_by_axis(
            slices=slices, volume_shape=volume_shape
        )
        _logger.debug("slices ordered by axis %s", indices_to_read_per_axis)

        for i_axis_0, file_name in enumerate(
            self.browse_data_files()  # pylint: disable=E1101
        ):
            # browse all files. Each file represent a slice along the axis 0
            data_cache = None
            # used to cache data in the case a full slice
            # needs to be read along axis 0. Then we can reuse it for over axis

            # handle axis 0
            if i_axis_0 in indices_to_read_per_axis[0]:
                data_cache = self.read_file(file_name=file_name)
                result[(0, i_axis_0)][:] = data_cache
                indices_to_read_per_axis[0].remove(i_axis_0)

            # test: if all slices are read avoid going through all the slices
            if numpy.sum([len(indices_to_read_per_axis[i]) for i in range(3)]) == 0:
                break

            # if data has been read already let's benefit from it
            if data_cache is not None:
                for index in indices_to_read_per_axis[1]:
                    result[(1, index)][i_axis_0:] = data_cache[index, :]
                for index in indices_to_read_per_axis[2]:
                    result[(2, index)][i_axis_0, :] = data_cache[:, index]
            else:
                # else read lines and / or columns directly
                if len(indices_to_read_per_axis[1]) > 0:
                    data = self.read_n_columns_in_file(
                        file_name=file_name, column_indices=indices_to_read_per_axis[1]
                    )
                    assert len(data) == len(indices_to_read_per_axis[1])
                    for column, index in zip(data, indices_to_read_per_axis[1]):
                        result[(1, index)][i_axis_0, :] = column

                if len(indices_to_read_per_axis[2]) > 0:
                    data = self.read_n_lines_in_file(
                        file_name=file_name, line_indices=indices_to_read_per_axis[2]
                    )
                    assert len(data) == len(indices_to_read_per_axis[2])
                    for line, index in zip(data, indices_to_read_per_axis[2]):
                        result[(2, index)][i_axis_0, :] = line

        return result

    def read_n_lines_in_file(self, file_name, line_indices) -> tuple:
        raise NotImplementedError("Base class")

    def read_n_columns_in_file(self, file_name, column_indices) -> tuple:
        raise NotImplementedError("Base class")

    def read_file(self, file_name) -> tuple:
        raise NotImplementedError("Base class")



# coding: utf-8

import os

import pytest

from tomoscan.esrf.edfscan import EDFTomoScan
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan.esrf.identifier.hdf5Identifier import NXtomoScanIdentifier
from tomoscan.factory import Factory
from tomoscan.identifier import ScanIdentifier
from tomoscan.tests.utils import NXtomoMockContext


def test_hdf5_identifier(tmp_path):
    """insure identifier is working for hdf5"""
    test_dir = tmp_path / "test_dir"
    os.makedirs(test_dir)
    with NXtomoMockContext(
        scan_path=str(test_dir),
        n_proj=10,
        n_ini_proj=10,
        distance=1.0,
        energy=1.0,
    ) as scan:
        identifier = scan.get_identifier()
        assert isinstance(identifier, ScanIdentifier)
        assert str(identifier).startswith(f"hdf5:{NXtomoScanIdentifier.TOMO_TYPE}:")
        scan_from_identifier = NXtomoScan.from_identifier(identifier)
        assert scan_from_identifier.entry == scan.entry
        assert os.path.abspath(scan_from_identifier.path) == os.path.abspath(scan.path)
        assert os.path.abspath(scan_from_identifier.master_file) == os.path.abspath(
            scan.master_file
        )

        scan_from_factory = Factory.create_tomo_object_from_identifier(
            identifier=str(identifier)
        )
        assert scan_from_factory.entry == scan.entry
        assert os.path.abspath(scan_from_factory.path) == os.path.abspath(scan.path)
        assert os.path.abspath(scan_from_factory.master_file) == os.path.abspath(
            scan.master_file
        )

        with pytest.raises(TypeError):
            EDFTomoScan.from_identifier(identifier)

        # insure if TOMO_TYPE is not provided then it will still be considered as a scan
        identifier_as_str = str(identifier)
        identifier_manual_def = ":".join(
            [identifier_as_str.split(":")[0], identifier_as_str.split(":")[-1]]
        )
        scan_from_factory = Factory.create_tomo_object_from_identifier(
            identifier=str(identifier_manual_def)
        )
        assert scan_from_factory is not None
        assert isinstance(identifier.short_description(), str)
        assert identifier == identifier.from_str(identifier.to_str())
        assert identifier != "toto"



from tomoscan.utils.io import deprecated_warning

deprecated_warning(
    "Module",
    name="tomoscan.serie",
    reason="Fix typo",
    replacement="tomoscan.series",
    only_once=True,
)

from .series import *  # noqa F401



from __future__ import annotations

import logging
import numpy
from .datasetscope import DatasetScope
from .method import Method

_logger = logging.getLogger(__name__)


class _DatasetInfos:
    def __init__(self):
        self._scope = DatasetScope.GLOBAL
        self._file_path = None
        self._data_path = None

    @property
    def scope(self) -> DatasetScope:
        return self._scope

    @scope.setter
    def scope(self, scope: str | DatasetScope):
        self._scope = DatasetScope(scope)

    @property
    def file_path(self):
        return self._file_path

    @file_path.setter
    def file_path(self, file_path):
        self._file_path = file_path

    @property
    def data_path(self):
        return self._data_path

    @data_path.setter
    def data_path(self, data_path: str):
        self._data_path = data_path


class _ROIInfo:
    def __init__(self, x_min=None, x_max=None, y_min=None, y_max=None):
        self.x_min = x_min
        self.x_max = x_max
        self.y_min = y_min
        self.y_max = y_max


class IntensityNormalization:
    """Information regarding the intensity normalization to be done"""

    def __init__(self):
        self._method = Method.NONE
        self._extra_info = {}

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method: str | Method | None):
        if method is None:
            method = Method.NONE
        self._method = Method(method)

    def set_extra_infos(self, info: dict | _DatasetInfos | _ROIInfo):
        if info is None:
            self._extra_info = None
        elif not isinstance(info, (_DatasetInfos, _ROIInfo, dict)):
            raise TypeError(
                "info is expected to be an instance of _DatasetInfos or _ROIInfo"
            )
        else:
            self._extra_info = info

    def get_extra_infos(self) -> dict | _DatasetInfos | _ROIInfo:
        return self._extra_info

    def to_dict(self) -> dict:
        res = {
            "method": self.method.value,
        }
        if self._extra_info not in (None, {}):
            res["extra_infos"] = self.get_extra_infos()
        return res

    def load_from_dict(self, dict_):
        if "method" in dict_:
            self.method = dict_["method"]
        if "extra_infos" in dict_:
            self.set_extra_infos(dict_["extra_infos"])
        return self

    @staticmethod
    def from_dict(dict_):
        res = IntensityNormalization()
        res.load_from_dict(dict_)
        return res

    def __str__(self):
        return f"method: {self.method}, extra-infos: {self.get_extra_infos()}"


def normalize_chebyshev_2D(sino):
    Nr, Nc = sino.shape
    J = numpy.arange(Nc)
    x = 2.0 * (J + 0.5 - Nc / 2) / Nc
    sum0 = Nc
    f2 = 3.0 * x * x - 1.0
    sum1 = (x**2).sum()
    sum2 = (f2**2).sum()
    for i in range(Nr):
        ff0 = sino[i, :].sum()
        ff1 = (x * sino[i, :]).sum()
        ff2 = (f2 * sino[i, :]).sum()
        sino[i, :] = sino[i, :] - (ff0 / sum0 + ff1 * x / sum1 + ff2 * f2 / sum2)
    return sino


def normalize_lsqr_spline_2D(sino):
    try:
        from scipy.interpolate import splev, splrep
    except ImportError:
        _logger.error("You should install scipy to do the lsqr spline normalization")
        return None

    Nr, Nc = sino.shape
    # correction = numpy.zeros_like(sino)
    for i in range(Nr):
        line = sino[i, :]
        spline = splrep(range(len(line)), sino[i, :], k=1)
        correct = splev(range(len(line)), spline)
        sino[i, :] = line - correct
    return sino



import numpy as np
from silx.io.utils import open as hdf5_open

__all__ = [
    "get_first_hdf5_entry",
    "get_h5_value",
    "get_h5obj_value",
    "get_hdf5_dataset_shape",
]


def get_first_hdf5_entry(fname):
    with hdf5_open(fname) as fid:
        entry = list(fid.keys())[0]
    return entry


def get_h5_value(fname, h5_path, default_ret=None):
    with hdf5_open(fname) as fid:
        try:
            val_ptr = fid[h5_path][()]
            # TODO: look at silx.io.utils.h5py_read_dataset might replace it. See https://gitlab.esrf.fr/tomotools/tomoscan/-/issues/88
        except KeyError:
            val_ptr = default_ret
    return val_ptr


def get_h5obj_value(h5_obj, name, default=None):
    if name in h5_obj:
        return h5_obj[name][()]
    return default


def _get_3D_subregion(sub_region):
    if sub_region is None:
        xmin, xmax, ymin, ymax, zmin, zmax = None, None, None, None, None, None
    elif len(sub_region) == 3:
        first_part, second_part, third_part = sub_region
        xmin, xmax = first_part
        ymin, ymax = second_part
        zmin, zmax = third_part
    elif len(sub_region) == 6:
        xmin, xmax, ymin, ymax, zmin, zmax = sub_region
    else:
        raise ValueError(
            "Expected parameter in the form (xmin, xmax, ymin, ymax, zmin, zmax) or ((xmin, xmax), (ymin, ymax), (zmin, zmax))"
        )
    return xmin, xmax, ymin, ymax, zmin, zmax


def get_hdf5_dataset_shape(fname, h5_data_path, sub_region=None):
    zmin, zmax, ymin, ymax, xmin, xmax = _get_3D_subregion(sub_region)
    with hdf5_open(fname) as f:
        d_ptr = f[h5_data_path]
        shape = d_ptr.shape
    n_z, n_y, n_x = shape
    # perhaps there is more elegant
    res_shape = []
    for n, bounds in zip([n_z, n_y, n_x], ((zmin, zmax), (ymin, ymax), (xmin, xmax))):
        res_shape.append(np.arange(n)[bounds[0] : bounds[1]].size)
    return tuple(res_shape)



# coding: utf-8

"""This module is dedicated to instances of :class:`BaseIdentifier` used at esrf"""

from .edfidentifier import EDFTomoScanIdentifier  # noqa F401
from .hdf5Identifier import NXtomoScanIdentifier  # noqa F401
from .jp2kidentifier import JP2KVolumeIdentifier  # noqa F401
from .rawidentifier import RawVolumeIdentifier  # noqa F401
from .tiffidentifier import MultiTiffVolumeIdentifier  # noqa F401
from .tiffidentifier import TIFFVolumeIdentifier  # noqa F401



"""
Test the integration of NXtomo 3.0 (and moving to McStas).
Especially that having an old Nxtomo or a recent one won't affect the reading.
"""

from __future__ import annotations

import os
import shutil
import pytest
import numpy

from tomoscan.tests.datasets import GitlabDataset
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from tomoscan._CoordinateSystem import CoordinateSystem
from nxtomo.utils.transformation import (
    DetXFlipTransformation,
    DetYFlipTransformation,
    DetZFlipTransformation,
)
from nxtomo.nxobject.nxtransformations import NXtransformations


@pytest.fixture
def nxtomo_beforce_McStas(tmp_path):
    """
    Return a dataset ('Atomium_S2_holo1_HT_010nm_500proj_0003.h5') converted **before** moving from McStas (nxtomomill <2.0 and nxtomo < 3.0)
    """
    src_file = GitlabDataset.get_dataset(
        "h5_datasets/coordinate_system/before_McStas_0000.nx"
    )
    dst_file = os.path.join(tmp_path, "before_McStas_0000.nx")
    shutil.copyfile(src_file, dst_file)
    return dst_file


@pytest.fixture
def nxtomo_after_McStas(tmp_path):
    """
    Return a dataset ('Atomium_S2_holo1_HT_010nm_500proj_0003.h5') converted **after** moving from McStas (nxtomomill == 2.0 and nxtomo == 3.0)
    """
    src_file = GitlabDataset.get_dataset(
        "h5_datasets/coordinate_system/after_McStas_0000.nx"
    )
    dst_file = os.path.join(tmp_path, "after_McStas_0000.nx")
    shutil.copyfile(src_file, dst_file)
    return dst_file


@pytest.mark.parametrize(
    "lr_flip, ud_flip", ((False, False), (True, False), (False, True))
)
def test_transition_to_McStas(
    nxtomo_beforce_McStas, nxtomo_after_McStas, lr_flip, ud_flip
):
    """
    Check that tomoscan API returns the same values before and after moving to McStas.
    This transition affects the NXtomo and NXtomomill but should affect tomoscan.

    Note on flip: the original dataset had no flip. We mock some in order to have a larger case coverage
    """
    # overwrite flips
    mc_stas_transformation = NXtransformations()
    mc_stas_transformation.add_transformation(DetYFlipTransformation(flip=lr_flip))
    mc_stas_transformation.add_transformation(DetXFlipTransformation(flip=ud_flip))
    mc_stas_transformation.save(
        file_path=nxtomo_after_McStas,
        data_path="/entry0000/instrument/detector/transformations",
        overwrite=True,
    )

    esrf_transformations = NXtransformations()
    esrf_transformations.add_transformation(DetZFlipTransformation(flip=lr_flip))
    esrf_transformations.add_transformation(DetYFlipTransformation(flip=ud_flip))
    esrf_transformations.save(
        file_path=nxtomo_beforce_McStas,
        data_path="/entry0000/instrument/detector/transformations",
        overwrite=True,
    )

    # create 'NXtomoScan' instances and test one against the other
    nxtomo_esrf_coord = NXtomoScan(nxtomo_beforce_McStas)
    nxtomo_mc_stas_coord = NXtomoScan(nxtomo_after_McStas)

    # check coordinate system
    assert (
        nxtomo_esrf_coord._get_coordinate_system(default=None) is CoordinateSystem.ESRF
    )
    assert (
        nxtomo_mc_stas_coord._get_coordinate_system(default=None)
        is CoordinateSystem.McStas
    )
    # check rotation angle
    # warning: this dataset has rotation defined 'counter clockwise'. So the value read by tomoscan should be untouched.
    # In nxtomomill < 2.0 this field was not interpreted and the use case of a counter-clockwise rotation was not considered but we should have invert them.

    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.rotation_angle,
        numpy.array(nxtomo_esrf_coord.rotation_angle),
    )
    # check translations
    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.x_translation, nxtomo_esrf_coord.x_translation
    )
    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.y_translation, nxtomo_esrf_coord.y_translation
    )
    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.z_translation, nxtomo_esrf_coord.z_translation
    )

    # check detector flips
    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.detector_is_lr_flip, nxtomo_esrf_coord.detector_is_lr_flip
    )
    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.detector_is_ud_flip, nxtomo_esrf_coord.detector_is_ud_flip
    )
    # check 'x_rotation_axis_pixel_position'
    numpy.testing.assert_almost_equal(
        nxtomo_mc_stas_coord.x_rotation_axis_pixel_position,
        nxtomo_esrf_coord.x_rotation_axis_pixel_position,
    )



from __future__ import annotations

import os
import h5py
import numpy
import fabio
import pint

from silx.io.url import DataUrl

from tomoscan.framereducer.reducedframesinfos import ReducedFramesInfos

_ureg = pint.get_application_registry()


class ReduceFrameSaver:
    """
    Util function to save reduced frames to disk.
    """

    def __init__(
        self,
        frames: dict[int, numpy.ndarray],
        output_urls: tuple[DataUrl, ...],
        frames_metadata: ReducedFramesInfos | None,
        metadata_output_urls: tuple | None,
        overwrite: bool = False,
    ):
        """
        :param frames: Frames as a dict to be saved. Key is the index of the frame, value is the frame as a 2D numpy array.
        :param output_urls: A list of URLs specifying the locations where the frames will be saved.
            Each URL in the list can include specific keywords that the frame saver will dynamically replace.
            These keywords allow for customizable naming conventions for the saved frames. Supported keywords are:

            - '{index}': This keyword will be replaced by the index of the frame.
            - '{index_zfill4}': This keyword will be replaced by the index of the frame, zero-padded to four digits.
        :param frames_metadata: Metadata associated to the frames (like count time).
        :param metadata_output_urls: A list of URLs specifying the locations where the metadata of the reduced frames will be saved.
        :param overwrite: If True and if one of the output url already exists it will overwrite it. Else it will raise an Error.
        """
        if not isinstance(frames, dict):
            raise TypeError(
                f"inputs `frames` is expected to be a dict not {type(frames)}"
            )
        if not isinstance(output_urls, (list, tuple, set)):
            raise TypeError(
                f"output_urls is expected to be a tuple not a {type(output_urls)}"
            )

        if frames_metadata is not None:
            if not isinstance(frames_metadata, ReducedFramesInfos):
                raise TypeError(
                    f"darks_infos is a {type(frames_metadata)} when None or {ReducedFramesInfos} expected"
                )
            self._check_reduced_infos(reduced_frames=frames, infos=frames_metadata)

        self.frames = frames
        self.output_urls = output_urls
        self.frames_metadata = frames_metadata
        self.metadata_output_urls = metadata_output_urls
        self.overwrite = overwrite

    @staticmethod
    def _check_reduced_infos(reduced_frames, infos):
        incoherent_metadata_mess = "incoherent provided infos:"
        incoherent_metadata = False
        if len(infos.count_time) not in (0, len(reduced_frames)):
            incoherent_metadata = True
            incoherent_metadata_mess += f"\n - count_time gets {len(infos.count_time)} when 0 or {len(reduced_frames)} expected"
        if len(infos.machine_current) not in (0, len(reduced_frames)):
            incoherent_metadata = True
            incoherent_metadata_mess += f"\n - machine_current gets {len(infos.machine_current)} when 0 or {len(reduced_frames)} expected"
        if incoherent_metadata:
            raise ValueError(incoherent_metadata_mess)

    def clean_frame_group(self, url):
        """
        For HDF5 in order to avoid file size increase we need to overwrite dataset when possible.
        But the darks / flats groups can contain other datasets and pollute this group.
        This function will remove unused dataset (frame index) when necessary
        """
        file_path = self._format_file_path(
            url.file_path(), index=None, index_zfill4=None
        )
        if not (os.path.exists(file_path) and h5py.is_hdf5(file_path)):
            return

        group_path = "/".join(
            self._format_data_path(url.data_path(), index=0, index_zfill4="0000").split(
                "/"
            )[:-1]
        )
        used_datasets = []
        for idx, _ in self.frames.items():
            idx_zfill4 = str(idx).zfill(4)
            used_datasets.append(
                self._format_data_path(
                    url.data_path(), index=idx, index_zfill4=idx_zfill4
                ).split("/")[-1]
            )
        with h5py.File(file_path, mode="a") as h5s:
            if group_path in h5s:
                if not self.overwrite:
                    raise KeyError("group_path already exists")
                for key in h5s[group_path].keys():
                    if key not in used_datasets:
                        del h5s[group_path][key]

    def save(self):
        # save data
        for url in self.output_urls:
            self.clean_frame_group(url=url)
            # first delete keys that are no more used
            for i_frame, (idx, frame) in enumerate(self.frames.items()):
                if not isinstance(frame, numpy.ndarray):
                    raise TypeError("frames are expected to be 2D numpy.ndarray")
                elif frame.ndim == 3 and frame.shape[0] == 1:
                    frame = frame.reshape([frame.shape[1], frame.shape[2]])
                elif frame.ndim != 2:
                    raise ValueError("frames are expected to be 2D numpy.ndarray")
                idx_zfill4 = str(idx).zfill(4)
                data_path = self._format_data_path(
                    url.data_path(), index=idx, index_zfill4=idx_zfill4
                )

            # small hack to insure 'flats' or 'darks' group are cleaned when start to write in
            for i_frame, (idx, frame) in enumerate(self.frames.items()):
                if not isinstance(frame, numpy.ndarray):
                    raise TypeError("frames are expected to be 2D numpy.ndarray")
                elif frame.ndim == 3 and frame.shape[0] == 1:
                    frame = frame.reshape([frame.shape[1], frame.shape[2]])
                elif frame.ndim != 2:
                    raise ValueError("frames are expected to be 2D numpy.ndarray")
                idx_zfill4 = str(idx).zfill(4)
                data_path = self._format_data_path(
                    url.data_path(), index=idx, index_zfill4=idx_zfill4
                )

                file_path = self._format_file_path(
                    url.file_path(), index=idx, index_zfill4=idx_zfill4
                )
                scheme = url.scheme()

                if scheme == "fabio":
                    if data_path is not None:
                        raise ValueError("fabio does not handle data_path")
                    else:
                        # for edf: add metadata to the header if some, without taking into account the
                        # metadata_output_urls (too complicated for backward compatibility...)
                        header = {}
                        if (
                            self.frames_metadata is not None
                            and len(self.frames_metadata.machine_current) > 0
                        ):
                            header["SRCUR"] = self.frames_metadata.machine_current[
                                i_frame
                            ]
                        if (
                            self.frames_metadata is not None
                            and len(self.frames_metadata.count_time) > 0
                        ):
                            header["CountTime"] = self.frames_metadata.count_time[
                                i_frame
                            ]

                        edf_writer = fabio.edfimage.EdfImage(
                            data=frame,
                            header=header,
                        )
                        edf_writer.write(file_path)
                elif scheme in ("hdf5", "silx"):
                    os.makedirs(os.path.dirname(file_path), exist_ok=True)
                    with h5py.File(file_path, mode="a") as h5s:
                        if data_path in h5s:
                            h5s[data_path][()] = frame
                        else:
                            h5s[data_path] = frame
                        h5s[data_path].attrs["interpretation"] = "image"
                else:
                    raise ValueError(
                        f"scheme {scheme} is not handled for frames. Should be fabio, silx of hdf5"
                    )

        frames_indexes = [idx for idx, _ in self.frames.items()]
        if self.frames_metadata is not None:
            for url, idx in zip(self.metadata_output_urls, frames_indexes):
                idx_zfill4 = str(idx).zfill(4)
                metadata_grp_path = self._format_data_path(
                    url.data_path(), index=idx, index_zfill4=idx_zfill4
                )
                file_path = self._format_file_path(
                    url.file_path(),
                    index=idx,
                    index_zfill4=idx_zfill4,
                )
                scheme = url.scheme()
                for (
                    metadata_name,
                    metadata_values,
                ) in self.frames_metadata.to_dict().items():
                    # warning: for now we only handle list (of count_time and machine_current)
                    if (not numpy.isscalar(metadata_values)) and len(
                        metadata_values
                    ) == 0:
                        continue
                    else:
                        # save metadata
                        if scheme in ("hdf5", "silx"):
                            with h5py.File(file_path, mode="a") as h5s:
                                metadata_path = "/".join(
                                    [metadata_grp_path, metadata_name]
                                )
                                if metadata_path in h5s:
                                    del h5s[metadata_path]
                                h5s[metadata_path] = metadata_values
                                unit = None
                                if metadata_name == ReducedFramesInfos.COUNT_TIME_KEY:
                                    unit = _ureg.second
                                elif (
                                    metadata_name
                                    == ReducedFramesInfos.MACHINE_ELECT_CURRENT_KEY
                                ):
                                    unit = _ureg.ampere
                                if unit is not None:
                                    h5s[metadata_path].attrs["units"] = str(unit)

                        else:
                            raise ValueError(
                                f"scheme {scheme} is not handled for frames metadata. Should be silx of hdf5"
                            )

    @staticmethod
    def _format_file_path(file_path: str | None, index, index_zfill4):
        if file_path is None:
            return file_path
        if not os.path.isabs(file_path):
            raise ValueError("url should contain file_path as an absolute path")
        return file_path.format(
            index=str(index),
            index_zfill4=index_zfill4,
        )

    @staticmethod
    def _format_data_path(data_path: str | None, index, index_zfill4):
        if data_path is None:
            return data_path
        return data_path.format(
            index_zfill4=index_zfill4,
            index=index,
        )



# coding: utf-8

"""module for giving information on process progress"""

import logging
import sys
from enum import Enum

_logger = logging.getLogger(__name__)

__all__ = [
    "Progress",
]


class _Advancement(Enum):
    step_1 = "\\"
    step_2 = "-"
    step_3 = "/"
    step_4 = "|"

    @staticmethod
    def getNextStep(step):
        if step is _Advancement.step_1:
            return _Advancement.step_2
        elif step is _Advancement.step_2:
            return _Advancement.step_3
        elif step is _Advancement.step_3:
            return _Advancement.step_4
        else:
            return _Advancement.step_1

    @staticmethod
    def getStep(value):
        if value % 4 == 0:
            return _Advancement.step_4
        elif value % 3 == 0:
            return _Advancement.step_3
        elif value % 2 == 0:
            return _Advancement.step_2
        else:
            return _Advancement.step_1


class Progress(object):
    """Simple interface for defining advancement on a 100 percentage base"""

    def __init__(self, name):
        self._name = name
        self.reset()

    def reset(self, max_=None):
        self._nProcessed = 0
        self._maxProcessed = max_

    def startProcess(self):
        self.setAdvancement(0)

    def setAdvancement(self, value):
        length = 20  # modify this to change the length
        block = int(round(length * value / 100))
        msg = f"\r{self._name}: [{'#' * block + '-' * (length - block)}] {round(value, 2)}%"
        if value >= 100:
            msg += " DONE\r\n"
        sys.stdout.write(msg)
        sys.stdout.flush()

    def endProcess(self):
        self.setAdvancement(100)

    def setMaxAdvancement(self, n):
        self._maxProcessed = n

    def increaseAdvancement(self, i=1):
        self._nProcessed += i
        self.setAdvancement((self._nProcessed / self._maxProcessed) * 100)



import os
import pytest
import numpy
from tomoscan.esrf.volume.helpers.single_frame_saver import SingleFrameSaverHelper
from tomoscan.esrf.volume.edfvolume import EDFVolume
from tomoscan.esrf.volume.jp2kvolume import JP2KVolume, has_glymur, has_minimal_openjpeg
from tomoscan.esrf.volume.tiffvolume import TIFFVolume, has_tifffile
from tomoscan.esrf.volume.mock import create_volume


volume_constructors = [
    EDFVolume,
]
if has_tifffile:
    volume_constructors.append(TIFFVolume)
if has_glymur and has_minimal_openjpeg:
    volume_constructors.append(JP2KVolume)


@pytest.mark.parametrize("cache_size", (None, 16, 256, 4096, 65536, numpy.inf))
@pytest.mark.parametrize("dtype", (numpy.uint8, numpy.uint16, numpy.float128))
@pytest.mark.parametrize("axis", (0, 1, 2))
@pytest.mark.parametrize("volume_constructor", volume_constructors)
def test_SingleFrameSaverHelper(tmp_path, axis, dtype, volume_constructor, cache_size):
    """Test SingleFrameSaverHelper along all the axis and with all kind of 'single frame' volume"""

    data = create_volume(frame_dims=(100, 100), z_size=100)
    # rescale data. Because for example jp2k is rescaling it by default
    if dtype is numpy.float128:
        max_val = numpy.finfo(dtype).max
    else:
        max_val = numpy.iinfo(dtype).max
    data = data / data.max() * max_val
    data = data.astype(dtype)

    output_folder = os.path.join(tmp_path, volume_constructor.DEFAULT_DATA_EXTENSION)
    volume = volume_constructor(
        folder=output_folder,
    )

    data_saver_helper = SingleFrameSaverHelper(
        volume=volume,
        data_shape=(100, 100, 100),
        dtype=dtype,
        cache_size=cache_size,
    )
    data_saver_helper.init_saver()

    for i in range(100):
        if axis == 0:
            data_saver_helper[i] = data[i]
        elif axis == 1:
            data_saver_helper[:, i] = data[:, i]
        elif axis == 2:
            data_saver_helper[:, :, i] = data[:, :, i]

    # note: saving is done directly on the file. So this will not affect the volume.data
    # except if loaded and stored
    volume_data = volume.load_data(store=False)

    assert data.shape == volume_data.shape
    assert data.dtype == volume_data.dtype

    numpy.testing.assert_array_equal(
        data,
        volume_data,
    )



"""Scan dedicated for bliss format - based on EDF files"""

from __future__ import annotations

import pint
import copy
import io
import json
import logging
import os
import re
import warnings

import fabio
import numpy
from lxml import etree
from silx.io.url import DataUrl


from tomoscan.utils.io import deprecated, deprecated_warning
from tomoscan.esrf.identifier.edfidentifier import EDFTomoScanIdentifier
from tomoscan.identifier import ScanIdentifier
from tomoscan.scanbase import Source, TomoScanBase
from tomoscan.utils import docstring

from .framereducer import EDFFrameReducer
from .utils import extract_urls_from_edf, get_parameters_frm_par_or_info

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)


__all__ = [
    "EDFTomoScan",
]


class EDFTomoScan(TomoScanBase):
    """
    TomoScanBase instanciation for scan defined from .edf files

    :param scan: path to the root folder containing the scan.
    :param dataset_basename: prefix of the dataset to handle
    :param scan_info: dictionary providing dataset information. Provided keys will overwrite information contained in .info.
                      Valid keys are: TODO
    :param n_frames: Number of frames in each EDF file.
        If not provided, it will be inferred by reading the files.
        In this case, the frame number is guessed from the file name.
    """

    _TYPE = "edf"

    INFO_EXT = ".info"

    ABORT_FILE = ".abo"

    _REFHST_PREFIX = "refHST"

    _DARKHST_PREFIX = "dark.edf"

    _SCHEME = "fabio"

    REDUCED_DARKS_DATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_darks.hdf5",
            data_path="{entry}/darks/{index}",
            scheme="silx",
        ),  # _darks.hdf5 and _flats.hdf5 are the default location of the reduced darks and flats.
        DataUrl(file_path="dark.edf", scheme=_SCHEME),
    )

    REDUCED_DARKS_METADATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_darks.hdf5",
            data_path="{entry}/darks/",
            scheme="silx",
        ),
        # even if no metadata urls are provided for EDF. If the output is the EDF metadata will be stored in the headers
    )

    REDUCED_FLATS_DATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_flats.hdf5",
            data_path="{entry}/flats/{index}",
            scheme="silx",
        ),  # _darks.hdf5 and _flats.hdf5 are the default location of the reduced darks and flats.
        DataUrl(
            file_path="refHST{index_zfill4}.edf", scheme=_SCHEME
        ),  # .edf is kept for compatiblity
    )

    REDUCED_FLATS_METADATAURLS = (
        DataUrl(
            file_path="{scan_prefix}_flats.hdf5",
            data_path="{entry}/flats/",
            scheme="silx",
        ),
        # even if no metadata urls are provided for EDF. If the output is the EDF metadata will be stored in the headers
    )

    FRAME_REDUCER_CLASS = EDFFrameReducer

    def __init__(
        self,
        scan: str | None,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
        n_frames: int | None = None,
        ignore_projections: dict | list | tuple | numpy.ndarray | None = None,
    ):
        TomoScanBase.__init__(
            self, scan=scan, type_=self._TYPE, ignore_projections=ignore_projections
        )

        # data caches
        self._darks = None
        self._flats = None
        self.__tomo_n = None
        self.__flat_n = None
        self.__dark_n = None
        self.__dim1 = None
        self.__dim2 = None
        self.__pixel_size = None
        self.__flat_on = None
        self.__scan_range = None
        self._edf_n_frames = n_frames
        self.__energy = None
        self._source = Source()
        """Source is not handle by EDFScan"""
        self._scan_info = None
        self.scan_info = scan_info
        self._dataset_basename = dataset_basename
        self._ignore_init = False
        # Used to handle ignored projections with EDF (done during reloading projections)
        self._projs_indices_and_angles = None

    def _rotation_angle(self):
        """
        Unofficial method for retrieving rotation angles.
        Can't do better without proper metadata.
        """
        if self.scan_range is None:
            raise ValueError("Need scan_range")
        fullturn = abs(self.scan_range - 360) < abs(self.scan_range - 180)
        angles = numpy.linspace(
            0, self.scan_range, num=self.tomo_n, endpoint=fullturn, dtype="f"
        )

        return angles

    def rotation_angle(self):
        if self.ignore_projections is None:
            return self._rotation_angle()
        else:
            all_indices, all_angles = self._projs_indices_and_angles
            ignored_indices = self.get_ignored_projection_indices()
            indices = sorted(list(set(all_indices) - set(ignored_indices)))
            return all_angles[indices]

    def _get_projs_indices_and_angles(self):
        # Caching is important, as projections indices will change once the ignore mechanism is applied!
        if self._projs_indices_and_angles is None:
            if self._projections is None:
                raise ValueError(
                    "Need _reload_projections (or evaluate self.projections)"
                )
            projs_indices = sorted(self._projections.keys())
            rot_angles = self._rotation_angle()  # heuristic, see above
            self._projs_indices_and_angles = projs_indices, rot_angles
        else:
            projs_indices, rot_angles = self._projs_indices_and_angles
        return projs_indices, rot_angles

    @property
    def scan_info(self) -> dict | None:
        return self._scan_info

    @scan_info.setter
    def scan_info(self, scan_info: dict | None) -> None:
        if not isinstance(scan_info, (type(None), dict)):
            raise TypeError("scan info is expected to be None or an instance of dict")
        used_keys = (
            "TOMO_N",
            "DARK_N",
            "REF_N",
            "REF_ON",
            "ScanRange",
            "Dim_1",
            "Dim_2",
            "Distance",
            "PixelSize",
            "SrCurrent",
        )
        other_keys = (
            "Prefix",
            "Directory",
            "Y_STEP",
            "Count_time",
            "Col_end",
            "Col_beg",
            "Row_end",
            "Row_beg",
            "Optic_used",
            "Date",
            "Scan_Type",
            "CCD_Mode",
            "CTAngle",
            "Min",
            "Max",
            "Sub_vols",
        )
        valid_keys = used_keys + other_keys
        valid_keys = [key.lower() for key in valid_keys]
        if isinstance(scan_info, dict):
            for key in scan_info.keys():
                if key not in valid_keys:
                    _logger.warning(f"{key} unrecognized. Valid keys are {valid_keys}")
        self._scan_info = scan_info

    @docstring(TomoScanBase.clear_cache)
    def clear_cache(self):
        super().clear_cache()
        self._projections = None
        self.__dim1 = None
        self.__dim2 = None
        self.__pixel_size = None

    def clear_frames_cache(self):
        self._darks = None
        self._flats = None
        self.__tomo_n = None
        self.__flat_n = None
        self.__dark_n = None
        self.__flat_on = None
        self.__scan_range = None
        super().clear_frames_cache()

    @docstring(TomoScanBase.tomo_n)
    @property
    def tomo_n(self) -> int | None:
        if self.__tomo_n is None:
            self.__tomo_n = EDFTomoScan.get_tomo_n(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__tomo_n

    @property
    @docstring(TomoScanBase.dark_n)
    def dark_n(self) -> int | None:
        if self.__dark_n is None:
            self.__dark_n = EDFTomoScan.get_dark_n(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__dark_n

    @property
    @docstring(TomoScanBase.flat_n)
    def flat_n(self) -> int | None:
        if self.__flat_n is None:
            self.__flat_n = EDFTomoScan.get_ref_n(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__flat_n

    @property
    @docstring(TomoScanBase.pixel_size)
    def pixel_size(self) -> int | None:
        """

        :return: pixel size
        """
        if self.__pixel_size is None:
            self.__pixel_size = EDFTomoScan._get_pixel_size(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__pixel_size

    @property
    def sample_x_pixel_size(self) -> float | None:
        """For EDF only square pixel size is handled"""
        return self.pixel_size

    @property
    def sample_y_pixel_size(self) -> float | None:
        """For EDF only square pixel size is handled"""
        return self.pixel_size

    @property
    def detector_x_pixel_size(self) -> float | None:
        # not provided by EDF at the moment
        return None

    @property
    def detector_y_pixel_size(self) -> float | None:
        # not provided by EDF at the moment
        return None

    @property
    @deprecated(replacement="", since_version="1.1.0")
    def x_real_pixel_size(self) -> float | None:
        if self.pixel_size is not None and self.magnification is not None:
            return self.pixel_size * self.magnification
        else:
            return None

    @property
    @deprecated(replacement="", since_version="1.1.0")
    def y_real_pixel_size(self) -> float | None:
        if self.pixel_size is not None and self.magnification is not None:
            return self.pixel_size * self.magnification
        else:
            return None

    @property
    @docstring(TomoScanBase.dim_1)
    def dim_1(self) -> int | None:
        """

        :return: image dim1
        """
        if self.__dim1 is None and self.path is not None:
            self.__dim1, self.__dim2 = EDFTomoScan.get_dim1_dim2(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__dim1

    @property
    @docstring(TomoScanBase.dim_2)
    def dim_2(self) -> int | None:
        """

        :return: image dim2
        """
        if self.__dim2 is None and self.path is not None:
            self.__dim1, self.__dim2 = EDFTomoScan.get_dim1_dim2(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__dim2

    @property
    @docstring(TomoScanBase.x_translation)
    def x_translation(self) -> tuple | None:
        _logger.warning("x_translation Not supported for EDF")
        return None

    @property
    @docstring(TomoScanBase.y_translation)
    def y_translation(self) -> tuple | None:
        _logger.warning("y_translation Not supported for EDF")
        return None

    @property
    @docstring(TomoScanBase.z_translation)
    def z_translation(self) -> tuple | None:
        _logger.warning("z_translation Not supported for EDF")
        return None

    @property
    @docstring(TomoScanBase.ff_interval)
    def ff_interval(self) -> int | None:
        if self.__flat_on is None and self.path is not None:
            self.__flat_on = EDFTomoScan.get_ff_interval(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__flat_on

    @property
    @docstring(TomoScanBase.scan_range)
    def scan_range(self) -> int | None:
        if self.__scan_range is None and self.path is not None:
            self.__scan_range = EDFTomoScan.get_scan_range(
                scan=self.path,
                dataset_basename=self.dataset_basename,
                scan_info=self.scan_info,
            )
        return self.__scan_range

    @property
    @docstring(TomoScanBase.flats)
    def flats(self) -> dict | None:
        """
        flats are given as a dictionary with index as key and DataUrl as
        value"""
        if self._flats is None and self.path is not None:
            self._flats = self.get_flats_url(
                scan_path=self.path,
                dataset_basename=self.dataset_basename,
            )
        return self._flats

    @property
    @docstring(TomoScanBase.projections)
    def projections(self) -> dict | None:
        if self._projections is None and self.path is not None:
            self._reload_projections()
        return self._projections

    @property
    @docstring(TomoScanBase.alignment_projections)
    def alignment_projections(self) -> None:
        if self._alignment_projections is None and self.path is not None:
            self._reload_projections()
        return self._alignment_projections

    @docstring(TomoScanBase.is_tomoscan_dir)
    @staticmethod
    def is_tomoscan_dir(
        directory: str, dataset_basename: str | None = None, **kwargs
    ) -> bool:
        return os.path.isfile(
            EDFTomoScan.get_info_file(
                directory=directory, dataset_basename=dataset_basename, kwargs=kwargs
            )
        )

    @staticmethod
    def get_info_file(
        directory: str, dataset_basename: str | None = None, **kwargs
    ) -> str:
        if dataset_basename is None:
            dataset_basename = os.path.basename(directory)
        assert dataset_basename != ""
        info_file = os.path.join(directory, dataset_basename + EDFTomoScan.INFO_EXT)

        if "src_pattern" in kwargs and kwargs["src_pattern"] is not None:
            assert "dest_pattern" in kwargs
            info_file = info_file.replace(
                kwargs["src_pattern"], kwargs["dest_pattern"], 1
            )
        return info_file

    @docstring(TomoScanBase.is_abort)
    def is_abort(self, **kwargs) -> bool:
        abort_file = self.dataset_basename + self.ABORT_FILE
        abort_file = os.path.join(self.path, abort_file)
        if "src_pattern" in kwargs and kwargs["src_pattern"] is not None:
            assert "dest_pattern" in kwargs
            abort_file = abort_file.replace(
                kwargs["src_pattern"], kwargs["dest_pattern"]
            )
        return os.path.isfile(abort_file)

    @property
    @docstring(TomoScanBase.darks)
    def darks(self) -> dict:
        if self._darks is None and self.path is not None:
            self._darks = self.get_darks_url(
                scan_path=self.path, dataset_basename=self.dataset_basename
            )
        return self._darks

    @docstring(TomoScanBase.get_proj_angle_url)
    def get_proj_angle_url(self) -> dict:
        # TODO: we might use fabio.open_serie instead
        if self.path is None:
            _logger.warning(
                "no path specified for scan, unable to retrieve the projections"
            )
            return {}
        n_projection = self.tomo_n
        data_urls = EDFTomoScan.get_proj_urls(
            self.path, dataset_basename=self.dataset_basename
        )
        return TomoScanBase.map_urls_on_scan_range(
            urls=data_urls, n_projection=n_projection, scan_range=self.scan_range
        )

    @docstring(TomoScanBase.update)
    def update(self):
        if self.path is not None:
            self._reload_projections()
            self._darks = EDFTomoScan.get_darks_url(self.path)
            self._flats = EDFTomoScan.get_flats_url(self.path)

    @docstring(TomoScanBase.load_from_dict)
    def load_from_dict(self, desc: dict | io.TextIOWrapper):
        if isinstance(desc, io.TextIOWrapper):
            data = json.load(desc)
        else:
            data = desc
        if not (self.DICT_TYPE_KEY in data and data[self.DICT_TYPE_KEY] == self._TYPE):
            raise ValueError("Description is not an EDFScan json description")

        assert self.DICT_PATH_KEY in data
        self.path = data[self.DICT_PATH_KEY]
        return self

    @staticmethod
    def get_proj_urls(
        scan: str,
        dataset_basename: str | None = None,
        n_frames: int | None = None,
    ) -> dict:
        """
        Return the dict of radios / projection for the given scan.
        Keys of the dictionary is the slice number
        Return all the file on the root of scan starting by the name of scan and
        ending by .edf

        :param scan: is the path to the folder of acquisition
        :param n_frames: Number of frames in each EDF file.
            If not provided, it is inferred by reading each file.
        :return: dict of radios files with radio index as key and file as value
        """
        urls = dict({})
        if (scan is None) or not (os.path.isdir(scan)):
            return urls
        if dataset_basename is None:
            dataset_basename = os.path.basename(scan)

        if os.path.isdir(scan):
            for f in os.listdir(scan):
                if EDFTomoScan.is_a_proj_path(
                    fileName=f, dataset_basename=dataset_basename, scanID=scan
                ):
                    gfile = os.path.join(scan, f)
                    index = EDFTomoScan.guess_index_frm_file_name(
                        gfile, basename=dataset_basename
                    )
                    urls.update(
                        extract_urls_from_edf(
                            start_index=index, file_=gfile, n_frames=n_frames
                        )
                    )

        return urls

    @staticmethod
    def is_a_proj_path(
        fileName: str, scanID: str, dataset_basename: str | None = None
    ) -> bool:
        """Return True if the given fileName can fit to a Radio name"""
        fileBasename = os.path.basename(fileName)
        if dataset_basename is None:
            dataset_basename = os.path.basename(scanID)

        if fileBasename.endswith(".edf") and fileBasename.startswith(dataset_basename):
            localstring = fileName.rstrip(".edf")
            # remove the scan
            localstring = re.sub(dataset_basename, "", localstring)
            if "slice_" in localstring:
                # case of a reconstructed file
                return False
            if "refHST" in localstring:
                return False
            s = localstring.split("_")
            if s[-1].isdigit():
                # check that the next value is a digit
                return True

        return False

    @staticmethod
    def guess_index_frm_file_name(_file: str, basename: str) -> int | None:
        """
        Guess the index of the file. Index is most of the an integer but can be
        a float for 'ref' for example if several are taken.

        :param _file:
        :param basename:
        """

        def extract_index(my_str, type_):
            res = []
            modified_str = copy.copy(my_str)
            while modified_str != "" and modified_str[-1].isdigit():
                res.append(modified_str[-1])
                modified_str = modified_str[:-1]
            if len(res) == 0:
                return None, modified_str
            else:
                orignalOrder = res[::-1]
                if type_ is int:
                    return int("".join(orignalOrder)), modified_str
                else:
                    return float(".".join(("0", "".join(orignalOrder)))), modified_str

        _file = os.path.basename(_file)
        if _file.endswith(".edf"):
            name = _file.replace(basename, "", 1)
            name = name.rstrip(".edf")

            part_1, name = extract_index(name, type_=int)
            if name.endswith("_"):
                name = name.rstrip("_")
                part_2, name = extract_index(name, type_=float)
            else:
                part_2 = None

            if part_1 is None:
                return None
            if part_2 is None:
                if part_1 is None:
                    return None
                else:
                    return int(part_1)
            else:
                return float(part_1) + part_2
        else:
            raise ValueError("only edf files are managed")

    @staticmethod
    def get_tomo_n(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> int | None:
        return EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="TOMO_N",
            type_=int,
            key_aliases=["tomo_N", "Tomo_N"],
            scan_info=scan_info,
        )

    @staticmethod
    def get_dark_n(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> int | None:
        return EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="DARK_N",
            type_=int,
            key_aliases=[
                "dark_N",
            ],
            scan_info=scan_info,
        )

    @staticmethod
    def get_ref_n(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> int | None:
        return EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="REF_N",
            type_=int,
            key_aliases=[
                "ref_N",
            ],
            scan_info=scan_info,
        )

    @staticmethod
    def get_ff_interval(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> int | None:
        return EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="REF_ON",
            type_=int,
            key_aliases=[
                "ref_On",
            ],
            scan_info=scan_info,
        )

    @staticmethod
    def get_scan_range(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> int | None:
        return EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="ScanRange",
            type_=int,
            key_aliases=[
                "scanRange",
            ],
            scan_info=scan_info,
        )

    @staticmethod
    def get_dim1_dim2(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> tuple | None:
        d1 = EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="Dim_1",
            key_aliases=["projectionSize/DIM_1"],
            type_=int,
            scan_info=scan_info,
        )
        d2 = EDFTomoScan.retrieve_information(
            scan=os.path.abspath(scan),
            dataset_basename=dataset_basename,
            ref_file=None,
            key="Dim_2",
            key_aliases=["projectionSize/DIM_2"],
            type_=int,
        )

        return d1, d2

    @property
    @docstring(TomoScanBase.instrument_name)
    def instrument_name(self) -> str | None:
        """

        :return: instrument name
        """
        return None

    @property
    @docstring(TomoScanBase.title)
    def title(self) -> str | None:
        """

        :return: title
        """
        return None

    @property
    @docstring(TomoScanBase.source_name)
    def source_name(self) -> str | None:
        """

        :return: source name
        """
        return None

    @property
    @docstring(TomoScanBase.source_type)
    def source_type(self) -> str | None:
        """

        :return: source type
        """
        return None

    @property
    def propagation_distance(self) -> float | None:
        """Not handled by EDF"""
        return None

    @property
    @docstring(TomoScanBase.sample_detector_distance)
    def sample_detector_distance(self) -> float | None:
        """sample / detector distance in meter"""
        if self._sample_detector_distance is None:
            self._sample_detector_distance = EDFTomoScan.retrieve_information(
                self.path,
                dataset_basename=self.dataset_basename,
                ref_file=None,
                key="Distance",
                type_=float,
                key_aliases=("distance",),
                scan_info=self.scan_info,
            )
        if self._sample_detector_distance is None:
            return None
        else:
            return (
                (self._sample_detector_distance * _ureg.millimeter)
                .to(_ureg.meter)
                .magnitude
            )

    @property
    def source_sample_distance(self) -> float | None:
        """Not handled for EDF"""
        return None

    @property
    @docstring(TomoScanBase.field_of_view)
    def field_of_view(self):
        # not managed for EDF files
        return None

    @property
    @docstring(TomoScanBase.x_rotation_axis_pixel_position)
    def x_rotation_axis_pixel_position(self):
        # not managed for EDF files
        return None

    @property
    @docstring(TomoScanBase.energy)
    def energy(self):
        if self.__energy is None:
            self.__energy = EDFTomoScan.retrieve_information(
                self.path,
                dataset_basename=self.dataset_basename,
                ref_file=None,
                key="Energy",
                type_=float,
                key_aliases=("energy",),
                scan_info=self.scan_info,
            )
        return self.__energy

    @property
    def count_time(self) -> list | None:
        if self._count_time is None:
            count_time = EDFTomoScan.retrieve_information(
                self.path,
                dataset_basename=self.dataset_basename,
                ref_file=None,
                key="Count_time",
                type_=float,
                key_aliases=("CountTime"),
                scan_info=self.scan_info,
            )
            if count_time is not None:
                if self.tomo_n is not None:
                    self._count_time = [count_time] * self.tomo_n
                else:
                    self._count_time = count_time

        return self._count_time

    @property
    def electric_current(self) -> tuple:
        warnings.warn(
            "electric_current is deprecated and will be removed in a future version. Use machine_current instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.machine_current

    @property
    @docstring(TomoScanBase.machine_current)
    def machine_current(self) -> tuple:
        if self._machine_current is None:
            machine_current = EDFTomoScan.retrieve_information(
                self.path,
                dataset_basename=self.dataset_basename,
                ref_file=None,
                key="SrCurrent",
                type_=float,
                key_aliases=("SRCUR", "machineCurrentStart"),
                scan_info=self.scan_info,
            )
            if machine_current is not None:
                if self.tomo_n is not None:
                    self._machine_current = [machine_current] * self.tomo_n
                else:
                    self._machine_current = machine_current

        return self._machine_current

    @staticmethod
    def _get_pixel_size(
        scan: str,
        dataset_basename: str | None = None,
        scan_info: dict | None = None,
    ) -> float | None:
        if os.path.isdir(scan) is False:
            return None
        value = EDFTomoScan.retrieve_information(
            scan=scan,
            dataset_basename=dataset_basename,
            ref_file=None,
            key="PixelSize",
            type_=float,
            key_aliases=[
                "pixelSize",
            ],
            scan_info=scan_info,
        )
        if value is None:
            parFile = os.path.join(scan, scan.dataset_basename + ".par")
            if os.path.exists(parFile):
                try:
                    ddict = get_parameters_frm_par_or_info(parFile)
                except ValueError as e:
                    _logger.error(e)
                if "IMAGE_PIXEL_SIZE_1".lower() in ddict:
                    value = float(ddict["IMAGE_PIXEL_SIZE_1".lower()])
        # for now pixel size are stored in microns.
        # We want to return them in meter
        if value is not None:
            return (value * _ureg.micrometer).to(_ureg.meter).magnitude
        else:
            return None

    @staticmethod
    def get_darks_url(
        scan_path: str,
        dataset_basename: str | None = None,
        prefix: str = "dark",
        file_ext: str = ".edf",
    ) -> dict:
        """

        :param scan_path:
        :param prefix: flat file prefix
        :param file_ext: flat file extension
        :return: list of flat frames as silx's `DataUrl`
        """
        res = {}
        if os.path.isdir(scan_path) is False:
            _logger.error(
                scan_path + " is not a directory. Cannot extract " "DarkHST files"
            )
            return res
        if dataset_basename is None:
            dataset_basename = os.path.basename(scan_path)
        for file_ in os.listdir(scan_path):
            _prefix = prefix
            if prefix.endswith(file_ext):
                _prefix = prefix.rstrip(file_ext)
            if file_.startswith(_prefix) and file_.endswith(file_ext):
                # usuelly the dark file name should be dark.edf, but some
                # darkHSTXXXX remains...
                file_fp = file_.lstrip(_prefix).rstrip(file_ext).lstrip("HST")
                if file_fp == "" or file_fp.isnumeric() is True:
                    index = EDFTomoScan.guess_index_frm_file_name(
                        _file=file_, basename=dataset_basename
                    )
                    urls = extract_urls_from_edf(
                        os.path.join(scan_path, file_), start_index=index
                    )
                    res.update(urls)
        return res

    @staticmethod
    def get_flats_url(
        scan_path: str,
        dataset_basename: str | None = None,
        prefix: str = "refHST",
        file_ext: str = ".edf",
        ignore=None,
    ) -> dict:
        """

        :param scan_path:
        :param prefix: flat frame file prefix
        :param file_ext: flat frame file extension
        :return: list of refs as silx's `DataUrl`
        """
        res = {}
        if os.path.isdir(scan_path) is False:
            _logger.error(
                scan_path + " is not a directory. Cannot extract " "RefHST files"
            )
            return res

        def get_next_free_index(key, keys):
            """return next free key from keys by converting it to a string
            with `key_value (n)` after it
            """
            new_key = key
            index = 2
            while new_key in keys:
                new_key = f"{key} ({index})"
                index += 1
            return new_key

        def ignore_file(file_name, to_ignore):
            if to_ignore is None:
                return False
            for pattern in to_ignore:
                if pattern in file_name:
                    return True
            return False

        if dataset_basename is None:
            dataset_basename = os.path.basename(scan_path)

        for file_ in os.listdir(scan_path):
            if (
                file_.startswith(prefix)
                and file_.endswith(file_ext)
                and not ignore_file(file_, ignore)
            ):
                index = EDFTomoScan.guess_index_frm_file_name(
                    _file=file_,
                    basename=dataset_basename,
                )
                file_fp = os.path.join(scan_path, file_)
                urls = extract_urls_from_edf(start_index=index, file_=file_fp)
                for key in urls:
                    if key in res:
                        key_ = get_next_free_index(key, res.keys())
                    else:
                        key_ = key
                    res[key_] = urls[key]
        return res

    @property
    def x_flipped(self) -> bool:
        deprecated_warning(
            type_="property",
            name="x_flipped",
            replacement="get_detector_transformations",
            since_version="1.3",
        )
        return None

    @property
    def y_flipped(self) -> bool:
        deprecated_warning(
            type_="property",
            name="y_flipped",
            replacement="detector_transformations",
            since_version="1.3",
        )
        return None

    @property
    def detector_transformations(self) -> tuple | None:
        """
        not handled for EDF
        """
        return None

    def _reload_projections(self):
        if self.path is None:
            return
        all_projections = EDFTomoScan.get_proj_urls(
            self.path,
            n_frames=self._edf_n_frames,
            dataset_basename=self.dataset_basename,
        )

        def select_proj(ddict, from_, to_):
            indexes = sorted(set(ddict.keys()))
            sel_indexes = indexes[from_:to_]
            res = {}
            for index in sel_indexes:
                res[index] = ddict[index]
            return res

        if self.tomo_n is not None and len(all_projections) > self.tomo_n:
            self._projections = select_proj(all_projections, 0, self.tomo_n)
            self._alignment_projections = select_proj(
                all_projections, self.tomo_n, None
            )
        else:
            self._projections = all_projections
            self._alignment_projections = {}
        if self._ignore_init is False:
            for idx in self.get_ignored_projection_indices():
                self._projections.pop(idx, None)
            self._ignore_init = True

    @staticmethod
    def retrieve_information(
        scan: str,
        dataset_basename: str | None,
        ref_file: str | None,
        key: str,
        type_: type,
        key_aliases: list | tuple | None = None,
        scan_info: dict | None = None,
    ):
        """
        Try to retrieve information a .info file, an .xml or a flat field file.

        file.
        Look for the key 'key' or one of it aliases.

        :param scan: root folder of an acquisition. Must be an absolute path
        :param ref_file: the refXXXX_YYYY which should contain information
                         about the scan. Ref in esrf reference is a flat.
        :param key: the key (information) we are looking for
        :param type_: required out type if the information is found
        :param key_aliases: aliases of the key in the different file
        :param scan_info: dict containing keys that could overwrite .info file content
        :return: the requested information or None if not found
        """
        info_aliases = [key]
        if key_aliases is not None:
            assert type(key_aliases) in (tuple, list)
            [info_aliases.append(alias) for alias in key_aliases]

        # if user provided 'scan_info' then he might want ot overwrite the metadata value
        if scan_info is not None:
            info = scan_info.get(key, scan_info.get(key.lower(), None))
            if info is not None:
                return info

        if not os.path.isdir(scan):
            return None

        # 1st look for ref file if any given
        def parseRefFile(filePath):
            with fabio.open(filePath) as ref_file:
                header = ref_file.header
            for k in key_aliases:
                if k in header:
                    return type_(header[k])
            return None

        if ref_file is not None and os.path.isfile(ref_file):
            try:
                info = parseRefFile(ref_file)
            except IOError as e:
                _logger.warning(e)
            else:
                if info is not None:
                    return info

        # 2nd look for .info file
        def parseInfoFile(filePath):
            metadata = get_parameters_frm_par_or_info(filePath)
            for alias in info_aliases:
                if alias.lower() in metadata:
                    return type_(metadata[alias.lower()])
            return None

        if dataset_basename is None:
            dataset_basename = os.path.basename(scan)
        infoFiles = [os.path.join(scan, dataset_basename + ".info")]
        infoOnDataVisitor = infoFiles[0].replace("lbsram", "")
        # Note: hack to check in /data/dataset instead of /lbsram/data.
        # when data is saved to /lbsram with spec the dataset is saved in /lbsram/data/{dataset} and the .info in /data/{dataset}
        if os.path.isfile(infoOnDataVisitor):
            infoFiles.append(infoOnDataVisitor)
        for infoFile in infoFiles:
            if os.path.isfile(infoFile) is True:
                info = parseInfoFile(infoFile)
                if info is not None:
                    return info

        # 3td look for xml files
        def parseXMLFile(filePath):
            try:
                for alias in info_aliases:
                    tree = etree.parse(filePath)
                    elmt = tree.find("acquisition/" + alias)
                    if elmt is None:
                        continue
                    else:
                        info = type_(elmt.text)
                        if info == -1:
                            return None
                        else:
                            return info
            except etree.XMLSyntaxError as e:
                _logger.warning(e)
                return None

        xmlFiles = [os.path.join(scan, dataset_basename + ".xml")]
        xmlOnDataVisitor = xmlFiles[0].replace("lbsram", "")
        # hack to check in lbsram, would need to be removed to add some consistency
        if os.path.isfile(xmlOnDataVisitor):
            xmlFiles.append(xmlOnDataVisitor)
        for xmlFile in xmlFiles:
            if os.path.isfile(xmlFile) is True:
                info = parseXMLFile(xmlFile)
                if info is not None:
                    return info

        return None

    def get_range(self):
        if self.path is not None:
            return self.get_scan_range(self.path, self.scan_info)
        else:
            return None

    def get_flat_expected_location(self):
        return os.path.join(self.dataset_basename, "refHST[*].edf")

    def get_dark_expected_location(self):
        return os.path.join(self.dataset_basename, "dark[*].edf")

    def get_projection_expected_location(self):
        return os.path.join(os.path.basename(self.path), self.dataset_basename, "*.edf")

    def _get_info_file_path_short_name(self):
        info_file = self.get_info_file_path(scan=self)
        return os.path.join(
            os.path.basename(os.path.dirname(info_file)), self.dataset_basename
        )

    def get_energy_expected_location(self):
        return "::".join((self._get_info_file_path_short_name(), "Energy"))

    def get_sample_detector_distance_expected_location(self):
        return "::".join((self._get_info_file_path_short_name(), "Distance"))

    def get_pixel_size_expected_location(self):
        return "::".join((self._get_info_file_path_short_name(), "PixelSize"))

    @staticmethod
    def get_info_file_path(scan):
        if not isinstance(scan, EDFTomoScan):
            raise TypeError(f"{scan} is expected to be an {EDFTomoScan}")
        if scan.path is None:
            return None
        scan_path = os.path.abspath(scan.path)
        return os.path.join(scan_path, scan.dataset_basename + ".info")

    def __str__(self):
        return f" edf scan({os.path.basename(os.path.abspath(self.path))})"

    @docstring(TomoScanBase.get_relative_file)
    def get_relative_file(self, file_name: str, with_dataset_prefix=True) -> str | None:
        if self.path is not None:
            if with_dataset_prefix:
                basename = self.dataset_basename
                basename = "_".join((basename, file_name))
                return os.path.join(self.path, basename)
            else:
                return os.path.join(self.path, file_name)
        else:
            return None

    def get_dataset_basename(self) -> str:
        return self.dataset_basename

    @property
    def dataset_basename(self) -> str | None:
        if self._dataset_basename is not None:
            return self._dataset_basename
        elif self.path is None:
            return None
        else:
            return os.path.basename(self.path)

    @docstring(TomoScanBase)
    def save_reduced_darks(
        self,
        darks: dict,
        output_urls: tuple = REDUCED_DARKS_DATAURLS,
        darks_infos=None,
        metadata_output_urls=REDUCED_DARKS_METADATAURLS,
        overwrite: bool = False,
    ):
        if len(darks) > 1:
            _logger.warning(
                "EDFTomoScan expect at most one dark. Only one will be save"
            )
        super().save_reduced_darks(
            darks=darks,
            output_urls=output_urls,
            darks_infos=darks_infos,
            metadata_output_urls=metadata_output_urls,
            overwrite=overwrite,
        )

    @docstring(TomoScanBase)
    def load_reduced_darks(
        self,
        inputs_urls: tuple = REDUCED_DARKS_DATAURLS,
        metadata_input_urls: tuple = REDUCED_DARKS_METADATAURLS,
        return_as_url: bool = False,
        return_info: bool = False,
    ) -> dict:
        darks = super().load_reduced_darks(
            inputs_urls=inputs_urls,
            metadata_input_urls=metadata_input_urls,
            return_as_url=return_as_url,
            return_info=return_info,
        )
        if return_info is True:
            darks, info = darks
        else:
            info = None
        # for edf we don't expect dark to have a index and we set it by default at frame index 0
        if None in darks:
            dark_frame = darks[None]
            del darks[None]
            if 0 in darks:
                _logger.warning("Two frame found for index 0")
            else:
                darks[0] = dark_frame
        if return_info is True:
            return darks, info
        else:
            return darks

    @docstring(TomoScanBase)
    def save_reduced_flats(
        self,
        flats: dict,
        output_urls: tuple = REDUCED_FLATS_DATAURLS,
        flats_infos=None,
        metadata_output_urls=REDUCED_FLATS_METADATAURLS,
        overwrite: bool = False,
    ) -> dict:
        super().save_reduced_flats(
            flats=flats,
            output_urls=output_urls,
            flats_infos=flats_infos,
            metadata_output_urls=metadata_output_urls,
            overwrite=overwrite,
        )

    @docstring(TomoScanBase)
    def load_reduced_flats(
        self,
        inputs_urls: tuple = REDUCED_FLATS_DATAURLS,
        metadata_input_urls: tuple = REDUCED_FLATS_METADATAURLS,
        return_as_url: bool = False,
        return_info=False,
    ) -> dict:
        return super().load_reduced_flats(
            inputs_urls=inputs_urls,
            return_as_url=return_as_url,
            return_info=return_info,
            metadata_input_urls=metadata_input_urls,
        )

    @docstring(TomoScanBase.compute_reduced_flats)
    def compute_reduced_flats(
        self,
        reduced_method="median",
        overwrite=True,
        output_dtype=numpy.int32,
        return_info=False,
    ):
        return super().compute_reduced_flats(
            reduced_method=reduced_method,
            overwrite=overwrite,
            output_dtype=output_dtype,
            return_info=return_info,
        )

    @docstring(TomoScanBase.compute_reduced_flats)
    def compute_reduced_darks(
        self,
        reduced_method="mean",
        overwrite=True,
        output_dtype=numpy.uint16,
        return_info=False,
    ):
        return super().compute_reduced_darks(
            reduced_method=reduced_method,
            overwrite=overwrite,
            output_dtype=output_dtype,
            return_info=return_info,
        )

    @staticmethod
    def from_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, EDFTomoScanIdentifier):
            raise TypeError(
                f"identifier should be an instance of {EDFTomoScanIdentifier} not {type(identifier)}"
            )
        return EDFTomoScan(scan=identifier.folder)

    @docstring(TomoScanBase)
    def get_identifier(self) -> ScanIdentifier:
        return EDFTomoScanIdentifier(
            object=self, folder=self.path, file_prefix=self.dataset_basename
        )



# coding: utf-8

"""test of the tomoscan.tomoobject module"""


import pytest

from tomoscan.tomoobject import TomoObject


def test_tomoobject():
    obj = TomoObject()
    with pytest.raises(NotImplementedError):
        obj.from_identifier("test")

    with pytest.raises(NotImplementedError):
        obj.get_identifier()



import contextlib
from os import path
import h5py
from silx.io.url import DataUrl
from silx.io.utils import open as hdf5_open


class _BaseReader(contextlib.AbstractContextManager):
    def __init__(self, url: DataUrl):
        if not isinstance(url, DataUrl):
            raise TypeError(f"url should be an instance of DataUrl. Not {type(url)}")
        if url.scheme() not in ("silx", "h5py"):
            raise ValueError("Valid scheme are silx and h5py")
        if url.data_slice() is not None:
            raise ValueError(
                "Data slices are not managed. Data path should "
                "point to a bliss node (h5py.Group)"
            )
        self._url = url
        self._file_handler = None

    def __exit__(self, *exc):
        return self._file_handler.close()


class DatasetReader(_BaseReader):
    """Context manager used to read a bliss node"""

    def __enter__(self):
        self._file_handler = hdf5_open(filename=self._url.file_path())
        entry = self._file_handler[self._url.data_path()]
        if not isinstance(entry, h5py.Dataset):
            raise ValueError(
                f"Data path ({self._url.path()}) should point to a dataset (h5py.Dataset)"
            )
        return entry


def get_data_sources(fname: str, data_path: str, recursive=True) -> dict:
    """
    Return the nested dict of data_sources.
     Key are (source) file path and keys are identical dict or data_path to the source

    :param recursive: If True then go trhough all linked files else won't 'solve' link
    """
    with hdf5_open(fname) as f:
        if data_path not in f:
            return None
        dptr = f[data_path]
        if not dptr.is_virtual:
            return {fname: data_path}
        sources = {}
        for vsource in dptr.virtual_sources():
            vsource_fname = path.normpath(
                path.join(path.dirname(dptr.file.filename), vsource.file_name)
            )
            if not path.isfile(vsource_fname):
                sources[vsource_fname] = None
            if recursive:
                sources[path.normpath(vsource_fname)] = get_data_sources(
                    vsource_fname, vsource.dset_name
                )
            else:
                sources[path.normpath(vsource_fname)] = vsource.dset_name

    return sources



# coding: utf-8

from tomoscan.esrf.identifier.rawidentifier import RawVolumeIdentifier


def test_vol_identifier():
    """insure identifier is working for .vol files"""
    from tomoscan.esrf.volume.rawvolume import RawVolume

    identifier = RawVolumeIdentifier(
        object=RawVolume,
        file_path="/dsad/test.vol",
    )
    assert str(identifier).startswith(f"raw:{RawVolumeIdentifier.TOMO_TYPE}:")



"""specific test for the jp2kvolume. the large part is in test_single_frame_volume as most of the processing is common"""

import os
import numpy
import pytest

from tomoscan.esrf.volume.jp2kvolume import JP2KVolume
from tomoscan.esrf.volume.mock import create_volume


@pytest.fixture
def raw_data(dtype: numpy.dtype):
    data = create_volume(
        frame_dims=(100, 100), z_size=11
    )  # z_size need to be at least 10 to check loading from file name works
    for i in range(len(data)):
        data[i] += 1
    return data.astype(dtype)


@pytest.mark.parametrize("dtype", (numpy.uint8, numpy.uint16))
def test_jp2kvolume_rescale(raw_data, dtype, tmp_path):
    """
    Test that rescale is correctly applied by default by the JP2KVolume
    """
    acquisition_dir = tmp_path / "acquisition"
    os.makedirs(acquisition_dir)
    volume_dir = str(acquisition_dir / "volume")
    os.makedirs(volume_dir)
    assert raw_data.dtype == dtype
    volume = JP2KVolume(folder=volume_dir, data=raw_data, metadata={})
    assert volume.data.dtype == dtype
    volume.save()

    volume.clear_cache()
    volume.load()
    assert volume.data.dtype == dtype
    assert volume.data.min() == 0
    assert volume.data.max() in (numpy.iinfo(dtype).max, numpy.iinfo(dtype).max - 1)
    assert volume.data_extension == "jp2"
    assert volume.metadata_extension == "txt"
